import sys
import os
import platform
from panda3d.core import WindowProperties
from panda3d.core import loadPrcFileData
from ursina.vec2 import Vec2
from ursina import color, application
from ursina.scene import instance as scene    # for toggling collider visibility
from ursina.string_utilities import print_info, print_warning
from screeninfo import get_monitors


class Window(WindowProperties):

    def ready(self, title, icon, borderless, fullscreen, size, forced_aspect_ratio, position, vsync, editor_ui_enabled, window_type, render_mode):
        loadPrcFileData('', 'window-title ursina')
        loadPrcFileData('', 'notify-level-util error')
        loadPrcFileData('', 'textures-auto-power-2 #t')
        loadPrcFileData('', 'load-file-type p3assimp')
        # loadPrcFileData('', 'allow-portal-cull #t')
        # loadPrcFileData("", "framebuffer-multisample 1")
        # loadPrcFileData('', 'multisamples 2')
        # loadPrcFileData('', 'textures-power-2 none')
        # loadPrcFileData('', 'threading-model Cull/Draw')
        loadPrcFileData('', 'coordinate-system y-up-left')
        # fallback to one of these if opengl is not supported
        loadPrcFileData('', 'aux-display pandadx9')
        loadPrcFileData('', 'aux-display pandadx8')
        loadPrcFileData('', 'aux-display tinydisplay')

        loadPrcFileData('', f'undecorated {borderless}')
        self.title = title
        self.icon = icon

        self.monitors = []
        self.main_monitor = None
        self.monitor_index = 0
        try:
            self.monitors = get_monitors()
            self.main_monitor = [e for e in self.monitors if e.is_primary][0]
            self.monitor_index = self.monitors.index(self.main_monitor)
        except:
            print_warning('no monitors found')

        if not size:
            if self.main_monitor:
                self.fullscreen_size = Vec2(self.main_monitor.width, self.main_monitor.height)
                self.windowed_size = self.fullscreen_size / 1.25
            else:
                self.fullscreen_size = Vec2(1280,720)
                self.windowed_size = self.fullscreen_size / 1.25
        else:
            self.fullscreen_size = size
            self.windowed_size = size

        if not fullscreen:
            loadPrcFileData('', f'win-size {self.windowed_size[0]} {self.windowed_size[1]}')
        else:
            loadPrcFileData('', f'win-size {self.fullscreen_size.x} {self.fullscreen_size.y}')

        self.windowed_position = None   # gets set when entering fullscreen so position will be correct when going back to windowed mode
        self.show_ursina_splash = False

        self.top = Vec2(0, .5)
        self.bottom = Vec2(0, -.5)
        self.center = Vec2(0, 0)


    def apply_settings(self):
        self.forced_aspect_ratio = None # example: window.forced_aspect_ratio = 16/9
        self.always_on_top = False

        self.vsync = True   # can't be set during play
        # self.borderless = borderless

        # if size:
        #     self.windowed_size = size
        # self.size = self.windowed_size

        # if fullscreen:
        #     self.fullscreen = fullscreen
        # self._fullscreen = not application.development_mode
        self.center_on_screen()



        self.color = color.dark_gray
        self.render_modes = ('default', 'wireframe', 'colliders', 'normals')
        self.render_mode = 'default'
        self.editor_ui = None
        if application.window_type != 'none':
            base.accept('aspectRatioChanged', self.update_aspect_ratio)
            if self.always_on_top:
                self.setZOrder(WindowProperties.Z_top)



    @property
    def left(self):
        return Vec2(-self.aspect_ratio/2, 0)
    @property
    def right(self):
        return Vec2(self.aspect_ratio/2, 0)
    @property
    def top_left(self):
        return Vec2(-self.aspect_ratio/2, .5)
    @property
    def top_right(self):
        return Vec2(self.aspect_ratio/2, .5)
    @property
    def bottom_left(self):
        return Vec2(-self.aspect_ratio/2, -.5)
    @property
    def bottom_right(self):
        return Vec2(self.aspect_ratio/2, -.5)


    def center_on_screen(self):
        if application.window_type == 'none':
            return
        x = self.main_monitor.x + ((self.main_monitor.width - self.size[0]) / 2)
        y = self.main_monitor.y + ((self.main_monitor.height - self.size[1]) / 2)
        self.position = Vec2(x,y)


    def make_editor_gui(self):     # called by main after setting up camera and application.development_mode
        from ursina import camera, Entity, Text, Button, ButtonList, Func, Tooltip, held_keys, mouse
        import time

        self.editor_ui = Entity(parent=camera.ui, eternal=True, enabled=bool(application.development_mode))

        def window_input(key):
            if key == 'f12':
                self.editor_ui.enabled = not self.editor_ui.enabled

            elif key == 'f11':
                self.fullscreen = not self.fullscreen

            elif key == 'f10':
                if held_keys['shift']:
                    self.render_mode = 'default'
                else:
                    i = self.render_modes.index(self.render_mode) + 1
                    if i >= len(self.render_modes):
                        i = 0

                    self.render_mode = self.render_modes[i]
        self.input_entity = Entity(name = 'window.input_entity', input=window_input)

        self.exit_button = Button(parent=self.editor_ui, text='x', eternal=True, ignore_paused=True, origin=(.5, .5), enabled=self.borderless,
            position=self.top_right, z=-999, scale=(.05, .025), color=color.red.tint(-.2), on_click=application.quit, name='exit_button')

        def _exit_button_input(key):
            if held_keys['shift'] and key == 'q' and not mouse.right:
                self.exit_button.on_click()
        self.exit_button.input = _exit_button_input

        self.fps_counter = Text(parent=self.editor_ui, eternal=True, text='60', ignore=False, i=0,
            position=((.5*self.aspect_ratio)-self.exit_button.scale_x, .47+(.02*(not self.exit_button.enabled)), -999))

        def _fps_counter_update():
            if self.fps_counter.i > 60:
                self.fps_counter.text = str(int(1//time.dt))
                self.fps_counter.i = 0
            self.fps_counter.i += 1
        self.fps_counter.update = _fps_counter_update

        self.entity_counter = Text(parent=self.editor_ui, eternal=True, origin=(-.5,.5), text='00', ignore=False, t=0,
            position=((.5*self.aspect_ratio)-self.exit_button.scale_x, .425+(.02*(not self.exit_button.enabled)), -999))
        self.entity_counter.text_entity = Text(parent=self.entity_counter, text='entities:', origin=(-.5,-.75), scale=.4, add_to_scene_entities=False, eternal=True)
        def _entity_counter_update():
            if self.entity_counter.t > 1:
                self.entity_counter.text = str(max(0, len([e for e in scene.entities if e.model and e.enabled])-6))
                self.entity_counter.i = 0
            self.entity_counter.t += time.dt
        self.entity_counter.update = _entity_counter_update

        self.collider_counter = Text(parent=self.editor_ui, eternal=True, origin=(-.5,.5), text='00', ignore=False, t=.1,
            position=((.5*self.aspect_ratio)-self.exit_button.scale_x, .38+(.02*(not self.exit_button.enabled)), -999))
        self.collider_counter.text_entity = Text(parent=self.collider_counter, text='colliders:', origin=(-.5,-.75), scale=.4, add_to_scene_entities=False, eternal=True)
        def _collider_counter_update():
            if self.collider_counter.t > 1:
                self.collider_counter.text = str(max(0, len([e for e in scene.entities if e.collider and e.enabled])-4))
                self.collider_counter.i = 0
            self.collider_counter.t += time.dt
        self.collider_counter.update = _collider_counter_update


        import webbrowser
        self.cog_menu = ButtonList({
            # 'Build' : Func(print, ' '),
            'API Reference' : Func(webbrowser.open, 'https://www.ursinaengine.org/api_reference.html'),
            # 'Asset Store' : Func(webbrowser.open, 'https://itch.io/tools/tag-ursina'),
            'ursfx (Sound Effect Maker)' : lambda: exec('from ursina.prefabs import ursfx; ursfx.open_gui()'),
            # 'Open Scene Editor' : Func(print, ' '),
            'Change Render Mode <gray>[F10]<default>' : self.next_render_mode,
            'Reset Render Mode <gray>[Shift+F10]<default>' : Func(setattr, self, 'render_mode', 'default'),
            'Toggle Hotreloading <gray>[F9]<default>' : application.hot_reloader.toggle_hotreloading,
            'Reload Shaders <gray>[F7]<default>' : application.hot_reloader.reload_shaders,
            'Reload Models <gray>[F7]<default>' : application.hot_reloader.reload_models,
            'Reload Textures <gray>[F6]<default>' : application.hot_reloader.reload_textures,
            'Reload Code <gray>[F5]<default>' : application.hot_reloader.reload_code,
        },
            width=.4, scale=.75, x=(.5*self.aspect_ratio)-(.4*.75), enabled=False, eternal=True, name='cog_menu', z=-10, color=color.black90, ignore_paused=True,
        )
        self.cog_menu.on_click = self.cog_menu.disable
        # print(self.cog_menu.scale_y)
        # self.cog_menu.scale *= .75
        self.cog_menu.highlight.color = color.azure
        self.cog_button = Button(parent=self.editor_ui, eternal=True, model='quad', texture='cog', scale=.015, origin=(1,-1), position=self.bottom_right, ignore_paused=True, name='cog_button')
        self.cog_menu.y = self.cog_button.y + (self.cog_menu.bg.scale_y * self.cog_menu.scale_y) + Text.size
        info_text ='''This menu is not enabled in builds. To see how the app will look like in builds, do Ursina(development_mode=False), which will disable all editor ui and start the app in fullscreen. To disable only this menu, do window.cog_menu.enabled = False'''
        self.cog_menu.info = Button(parent=self.cog_menu, model='circle', text='<gray>?', scale=.025, origin=(.5,-.5), tooltip=Tooltip(info_text, scale=.75, origin=(-.5,-.5), eternal=True), eternal=True, name='cog_menu_info')
        self.cog_menu.info.text_entity.scale *= .75
        def _toggle_cog_menu():
            self.cog_menu.enabled = not self.cog_menu.enabled
        self.cog_button.on_click = _toggle_cog_menu
        # print('-----------', time.time() - t) # 0.04


    def update_aspect_ratio(self):
        if hasattr(self, 'prev_size'):
            self.prev_aspect_ratio = self.prev_size[0] / self.prev_size[1]
        else:
            self.prev_aspect_ratio = self.aspect_ratio

        self.prev_size = self.size

        from ursina import camera, window
        value = [int(e) for e in base.win.getSize()]
        camera.set_shader_input('window_size', value)

        print_info('changed aspect ratio:', round(self.prev_aspect_ratio, 3), '->', round(self.aspect_ratio, 3))

        camera.ui_lens.set_film_size(camera.ui_size * .5 * self.aspect_ratio, camera.ui_size * .5)
        for e in [e for e in scene.entities if e.parent == camera.ui] + self.editor_ui.children:
            e.x /= self.prev_aspect_ratio / self.aspect_ratio

        if camera.orthographic:
            camera.orthographic_lens.set_film_size(camera.fov * window.aspect_ratio, camera.fov)
            base.cam.node().set_lens(camera.orthographic_lens)


    @property
    def position(self):
        if application.window_type == 'none':
            return Vec2(0,0)

        return Vec2(*self.getOrigin())

    @position.setter
    def position(self, value):
        if application.window_type == 'none':
            return
        print('set window position:', value)
        self.setOrigin(int(value[0]), int(value[1]))
        if application.base and hasattr(application.base.win, 'request_properties'):
            application.base.win.request_properties(self)


    @property
    def size(self):
        try:
            return Vec2(*base.win.getSize())
        except:
            return Vec2(1280,720)

    @size.setter
    def size(self, value):
        if application.window_type == 'none' or not value:
            return

        print('---------------set size to:', value)
        if hasattr(self, '_forced_aspect_ratio') and self.forced_aspect_ratio:
            value = (value[1] * self.forced_aspect_ratio, value[1])

        self.setSize(int(value[0]), int(value[1]))
        if application.base and hasattr(application.base.win, 'request_properties'):
            application.base.win.request_properties(self)
        try:
            from ursina import camera
            camera.set_shader_input('window_size', value)
        except:
            pass # not initialized yet


    @property
    def aspect_ratio(self):
        try:
            w, h = base.win.getSize()
            return w / h
        except:
            return 16/9

    @property
    def forced_aspect_ratio(self):
        if not hasattr(self, '_forced_aspect_ratio'): return None
        return self._forced_aspect_ratio

    @forced_aspect_ratio.setter
    def forced_aspect_ratio(self, value):
        self._forced_aspect_ratio = value
        if not value:
            return

        self.size = self.size


    @property
    def render_mode(self):
        if not hasattr(self, '_render_mode'): return None
        return self._render_mode

    @render_mode.setter
    def render_mode(self, value):
        self._render_mode = value
        # print('render mode:', value)
        base.wireframeOff()

        # disable collision display mode
        if hasattr(self, 'original_colors'):
            for i, e in enumerate([e for e in scene.entities if hasattr(e, 'color')]):
                e.color = self.original_colors[i]
                if e.collider:
                    e.collider.visible = False

        for e in [e for e in scene.entities if e.model and e.alpha]:
            e.setShaderAuto()

        if value == 'wireframe':
            base.wireframeOn()

        elif value == 'colliders':
            self.original_colors = [e.color for e in scene.entities if hasattr(e, 'color')]
            for e in scene.entities:
                e.color = color.clear
                if e.collider:
                    # e.visible = False
                    e.collider.visible = True

        elif value == 'normals':
            from ursina.shaders import normals_shader
            for e in [e for e in scene.entities if e.model and e.alpha]:
                e.shader = normals_shader
                e.set_shader_input('transform_matrix', e.getNetTransform().getMat())


    def next_render_mode(self):
        i = self.render_modes.index(self.render_mode) + 1
        if i >= len(self.render_modes):
            i = 0

        self.render_mode = self.render_modes[i]


    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, value):
        self._title = value
        loadPrcFileData('', f'window-title {value}')


    @property
    def icon(self):
        return self._icon

    @icon.setter
    def icon(self, value):
        self._icon = value
        self.setIconFilename(value)


    @property
    def borderless(self):
        return getattr(self, '_borderless', True)

    @borderless.setter
    def borderless(self, value):
        self._borderless = value
        if application.window_type == 'none': return

        self.setUndecorated(value)
        if hasattr(self, 'exit_button'):
            self.exit_button.enabled = not value

        if application.base and hasattr(application.base.win, 'request_properties'):
            application.base.win.request_properties(self)


    @property
    def fullscreen(self):
        return getattr(self, '_fullscreen', True)


    @fullscreen.setter
    def fullscreen(self, value):
        # print('----------------------', value, self._fullscreen)
        self._fullscreen = value
        if application.window_type == 'none': return

        print('set fullscreen to', value)
        if value:
            print(self.main_monitor)
            self.windowed_position = self.position
            self.windowed_size = self.size
            self.position = Vec2(self.main_monitor.x, self.main_monitor.y)
            self.size = Vec2(self.main_monitor.width, self.main_monitor.height)
        else:
            self.size = self.windowed_size
            if self.windowed_position is not None:
                self.position = self.windowed_position
            else:
                self.center_on_screen()

        # self.setFullscreen(value)
        base.win.request_properties(self)
        # except:
        #     print_warning('failed to set fullscreen', value)
        #     pass

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        self._color = value
        if application.window_type == 'none': return

        base.camNode.get_display_region(0).get_window().set_clear_color(value)


    @property
    def vsync(self):
        return self._vsync

    @vsync.setter
    def vsync(self, value):
        self._vsync = value
        if not 'base' in sys.modules:     # set vsync/framerate before window opened
            if value == True or value == False:
                loadPrcFileData('', f'sync-video {value}')
            elif isinstance(value, int):
                loadPrcFileData('', 'clock-mode limited')
                loadPrcFileData('', f'clock-frame-rate {value}')
        else:
            from panda3d.core import ClockObject                      # set vsync/framerate in runtime
            if value == True:
                globalClock.setMode(ClockObject.MNormal)
            elif value == False:
                print_warning('error: disabling vsync during runtime is not yet implemented')

            elif isinstance(value, (int, float, complex)):
                globalClock.setMode(ClockObject.MLimited)
                globalClock.setFrameRate(int(value))


instance = Window()


if __name__ == '__main__':
    from ursina import *
    # application.development_mode = False
    # app = Ursina(borderless=True, forced_aspect_ratio=16/9)
    app = Ursina(
        title='Ursina',
        # borderless=False,
        # borderless=True,
        # fullscreen=True,
        # show_ursina_splash=True,
        # development_mode=False,
        # vsync = True
        )
    button_list = ButtonList(
        {
        'widow.position = Vec2(0,0)': Func(setattr, window, 'position', Vec2(0,0)),
        'widow.size = Vec2(512,512)': Func(setattr, window, 'size', Vec2(512,512)),
        'widow.center_on_screen()': window.center_on_screen,

        'widow.borderless = True': Func(setattr, window, 'borderless', True),
        'widow.borderless = False': Func(setattr, window, 'borderless', False),

        'widow.fullscreen = True': Func(setattr, window, 'fullscreen', True),
        'widow.fullscreen = False': Func(setattr, window, 'fullscreen', False),

        'widow.vsync = True': Func(setattr, window, 'vsync', True),
        'widow.vsync = False': Func(setattr, window, 'vsync', False),


        'application.base.win.request_properties(self)': Func(application.base.win.request_properties, window),

        }, y=0
    )
    startup_value = Text(y=.5,x=-.5)
    startup_value.text = f'''
        position: {window.position}
        size: {window.size}
        aspect_ratio: {window.aspect_ratio}
        window.main_monitor.width: {window.main_monitor.width}
        window.main_monitor.height: {window.main_monitor.height}

    '''

    position_text = Text(y=.5,)

    # def update():
    #     position_text.text = f'''
    #         position: {window.position}
    #         size: {window.size}
    #         aspect_ratio: {window.aspect_ratio}
    #         window.main_monitor.width: {window.main_monitor.width}
    #         window.main_monitor.height: {window.main_monitor.height}
    #
    #     '''

    def input(key):
        if key == 'space':
            # x = window.main_monitor.x + ((window.main_monitor.width - window.size[0]) / 2)
            # y = window.main_monitor.y + ((window.main_monitor.height - window.size[1]) / 2)
            # window.position = Vec2(x,y)
            window.center_on_screen()

    # window.borderless = False
    # window.fullscreen = True

    # window.size = window.size - Vec2(1,1)
    # window.monitor_index = 0
    # window.center_on_screen()
    # print('------------', window.monitors)
    # time.sleep(2)
    # window.forced_aspect_ratio = 1
    # window.vsync = 10
    # window.title = 'ursina'
    # window.borderless = False
    # window.fullscreen = False
    # window.fps_counter.enabled = False
    # window.cog_button.enabled = False
    # window.position =(0,837)

    # camera.orthographic = True
    # camera.fov = 2
    # Text(text='adoij', x=.1)
    # def input(key):
    #     if key == 'space':
    #         window.center_on_screen()

    # Entity(model='cube', color=color.green, collider='box', texture='shore')
    app.run()



from panda3d.core import Vec3 as PandaVec3
from ursina.vec2 import Vec2



class Vec3(PandaVec3):
    def __round__(self, decimals=4):
        return Vec3(*(round(e,decimals) for e in self))


    def __repr__(self):
        return super().__repr__().replace('LVector3f', 'Vec3')


    def __iadd__(self, value):
        if len(value) % 3 == 0:
            for i in range(0, len(value), 3):
                self.add_x(value[i])
                self.add_y(value[i+1])
                self.add_z(value[i+2])
                return self

        if len(value) % 2 == 0:
            for i in range(0, len(value), 2):
                self.add_x(value[i])
                self.add_y(value[i+1])
            return self


    def __add__(self, value):
        if len(value) == 3:
            return Vec3(self[0]+value[0], self[1]+value[1], self[2]+value[2])

        if len(value) == 2:
            return Vec3(self[0]+value[0], self[1]+value[1], self[2])


    def __sub__(self, value):
        if len(value) == 3:
            return Vec3(self[0]-value[0], self[1]-value[1], self[2]-value[2])

        if len(value) == 2:
            return Vec3(self[0]-value[0], self[1]-value[1], self[2])


    @property
    def x(self):
        return self[0]
    @x.setter
    def x(self, value):
        self[0] = value

    @property
    def y(self):
        return self[1]
    @y.setter
    def y(self, value):
        self[1] = value

    @property
    def z(self):
        return self[2]
    @z.setter
    def z(self, value):
        self[2] = value

    @property
    def xy(self):
        return Vec2(self.x, self.y)
    @xy.setter
    def xy(self, value):
        self[0] = value[0]
        self[1] = value[1]

    @property
    def xz(self):
        return Vec2(self.x, self.z)
    @xz.setter
    def xz(self, value):
        self[0] = value[0]
        self[2] = value[1]

    @property
    def yz(self):
        return Vec2(self.y, self.z)
    @yz.setter
    def yz(self, value):
        self[1] = value[0]
        self[2] = value[1]

    @property
    def X(self):
        return int(self.x)
    @property
    def Y(self):
        return int(self.y)
    @property
    def Z(self):
        return int(self.z)


    def __mul__(self, value):
        if isinstance(value, (int, float, complex)):
            return Vec3(*(e*value for e in self))

        return Vec3(self[0]*value[0], self[1]*value[1], self[2]*value[2])


    __rmul__ = __mul__


    def __truediv__(self, value):
        if isinstance(value, (int, float, complex)):
            return Vec3(*(e/value for e in self))

        return Vec3(self[0]/value[0], self[1]/value[1], self[2]/value[2])

Vec3.zero = Vec3(0,0,0)
Vec3.one = Vec3(1,1,1)
Vec3.right = Vec3(1,0,0)
Vec3.left = Vec3(-1,0,0)
Vec3.up = Vec3(0,1,0)
Vec3.down = Vec3(0,-1,0)
Vec3.forward = Vec3(0,0,1)
Vec3.back = Vec3(0,0,-1)


if __name__ == '__main__':
    a = Vec3(1,0,0) * 2
    a = Vec3(1,0,1) * Vec3(2,1,2)
    b = Vec3(1.252352324,0,1)
    b += Vec3(0,1)

    # test
    print(a)
    b.x += 2
    print(b.x)
    print('xy:', b.xy)
    print(round(b))
    print(round(b.xy))
    print('-----------', a * 2)
    print('-----------', 2 * a)



class HitInfo:
    __slots__ = ['hit', 'entity', 'point', 'world_point', 'distance', 'normal', 'world_normal', 'hits', 'entities']


    def __init__(self, **kwargs):

        self.hit = None
        self.entity = None
        self.point = None
        self.world_point = None
        self.distance = 9999
        self.normal = None
        self.world_normal = None
        self.hits = []
        self.entities = []

        for key, value in kwargs.items():
            setattr(self, key, value)


    def __bool__(self):
        return self.hit



import sys
from pathlib import Path
from panda3d.core import getModelPath
from ursina import string_utilities

paused = False
time_scale = 1
calculate_dt = True
sequences = []
trace_entity_definition = False # enable to set entity.line_definition
print_entity_definition = False

package_folder = Path(__file__).parent
asset_folder = Path(sys.argv[0]).parent
blender_paths = dict()

development_mode = True
dirs = [e.stem for e in asset_folder.parent.iterdir() if e.is_dir()]
if 'src' in dirs and 'python' in dirs:
    development_mode = False

window_type = 'onscreen'
show_ursina_splash = False
gltf_no_srgb = True

print_info = development_mode
print_warnings = True
raise_exception_on_missing_model = False
raise_exception_on_missing_texture = False

internal_models_folder = package_folder / 'models/'
internal_models_compressed_folder = package_folder / 'models_compressed/'
internal_prefabs_folder = package_folder / 'prefabs/'
internal_scripts_folder = package_folder / 'scripts/'
internal_textures_folder = package_folder / 'textures/'
internal_fonts_folder = package_folder / 'fonts/'
internal_audio_folder = package_folder / 'audio/'

scenes_folder = asset_folder / 'scenes/'
scripts_folder = asset_folder / 'scripts/'
fonts_folder = asset_folder / 'fonts/'

compressed_textures_folder = asset_folder / 'textures_compressed/'
compressed_models_folder = asset_folder / 'models_compressed/'

# fonts are loaded py panda3d, so add paths here
_model_path = getModelPath()
_model_path.append_path(str(internal_fonts_folder.resolve()))
_model_path.append_path(str(Path('C:/Windows/Fonts').resolve()))
_model_path.append_path(str(asset_folder.resolve()))


base = None             # this will be set once the Ursina() is created
hot_reloader = None     # will be set my main if development_mode

def pause():
    global paused
    paused = True

def resume():
    global paused
    paused = False

def quit():
    sys.exit()


def load_settings(path=asset_folder / 'settings.py'):
    if path.exists():
        with open(path) as f:
            try:
                # d = dict(locals(), **globals())
                # exec(f.read(), d, d)
                exec('from ursina import *\n' + f.read())
                string_utilities.print_info('loaded settings from settings.py successfully')
            except Exception as e:
                string_utilities.print_warning('warning: settings.py error:', e)



class Scene(Entity):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        Entity(parent=self, position=Vec3(2.1786,0.033,6.256), model='cube', origin=Vec3(0,-0.5,0), ignore=True)
        Entity(parent=self, position=Vec3(-0.9924,0.032,5.2352), model='cube', origin=Vec3(0,-0.5,0), ignore=True)
        Entity(parent=self, position=Vec3(1.2385,0.032,6.6089), model='cube', origin=Vec3(0,-0.5,0), ignore=True)
        Entity(parent=self, position=Vec3(0.1564,0.032,6.6415), model='cube', origin=Vec3(0,-0.5,0), ignore=True)
        Entity(parent=self, position=Vec3(-1.9521,0.032,6.7786), model='cube', origin=Vec3(0,-0.5,0), ignore=True)
        Entity(parent=self, position=Vec3(3.7492,0.032,7.5643), model='cube', origin=Vec3(0,-0.5,0), ignore=True)



from panda3d.core import Vec4 as PandaVec4


class Vec4(PandaVec4):

    def __round__(self, decimals=4):
        return Vec4(*(round(e,decimals) for e in self))


    def __repr__(self):
        return f'Vec4({self[0]}, {self[1]}, {self[2]}, {self[3]})'


    def __iadd__(self, value):
        if len(value) % 3 == 0:
            for i in range(0, len(value), 3):
                self.add_x(value[i])
                self.add_y(value[i+1])
                self.add_z(value[i+2])
                return self

        if len(value) % 2 == 0:
            for i in range(0, len(value), 2):
                self.add_x(value[i])
                self.add_y(value[i+1])
            return self


    def __add__(self, value):
        if len(value) == 4:
            return Vec4(self[0]+value[0], self[1]+value[1], self[2]+value[2], self[3]+value[3])

        if len(value) == 3:
            return Vec4(self[0]+value[0], self[1]+value[1], self[2]+value[2], self[3])

        elif len(value) == 2:
            return Vec4(self[0]+value[0], self[1]+value[1], self[2], self[3])



    def __mul__(self, value):
        if isinstance(value, (int, float, complex)):
            return Vec4(*(e*value for e in self))

        return Vec4(self[0]*value[0], self[1]*value[1], self[2]*value[2], self[3]*value[3])


    __rmul__ = __mul__


    def __truediv__(self, value):
        if isinstance(value, (int, float, complex)):
            return Vec4(*(e/value for e in self))

        return Vec4(self[0]/value[0], self[1]/value[1], self[2]/value[2], self[3]/value[3])


if __name__ == '__main__':
    a = Vec4(1,0,0,0) * 2
    a = Vec4(1,0,1,1) * Vec4(2,1,2,3)
    b = Vec4(1.252352324,0,1,.2)
    b += Vec4(0,1)

    # test
    print(a)
    print(round(b))
    print('-----------', a * 2)
    print('-----------', 2 * a)



from ursina import input_handler, held_keys, Entity
from panda3d.core import InputDevice


if __name__ == '__main__':
    from ursina import Ursina, Text, time, color
    app = Ursina()
    text_entity = Text()
    player = Entity(model='cube', color=color.azure)

    def update():
        player.x += held_keys['gamepad left stick x'] * time.dt * 5
        player.y += held_keys['gamepad left stick y'] * time.dt * 5
        text_entity.text = '\n'.join([f'{key}: {held_keys[key]}' for key in ('gamepad left trigger', 'gamepad right trigger', 'gamepad right stick x', 'gamepad right stick y')])

    def input(key):
        print('key:', key)

    app.run()


input_handler.gamepad = None
input_handler.gamepads = base.devices.getDevices(InputDevice.DeviceClass.gamepad)
if input_handler.gamepads:
    input_handler.gamepad = input_handler.gamepads[0]


for i, gamepad in enumerate(input_handler.gamepads):
    gamepad_name = 'gamepad'
    if i > 0:
        gamepad_name += f'_{i}'

    base.attachInputDevice(gamepad, prefix=gamepad_name)
    buttons = {
        'face_a' : 'a',
        'face_b' : 'b',
        'face_x' : 'x',
        'face_y' : 'y',
        'lstick' : 'left stick',
        'rstick' : 'right stick',
        'back' : 'back',
        'start' : 'start',
        'dpad_up' : 'dpad up',
        'dpad_down' : 'dpad down',
        'dpad_left' : 'dpad left',
        'dpad_right' : 'dpad right',
        'rshoulder' : 'right shoulder',
        'lshoulder' : 'left shoulder',
    }

    for original_name, new_name in buttons.items():
        base.accept(f'{gamepad_name}-{original_name}', base.input, extraArgs=[f'{gamepad_name} {new_name}'])
        base.accept(f'{gamepad_name}-{original_name}-up', base.input, extraArgs=[f'{gamepad_name} {new_name} up'])
        # print(original_name, new_name)

def update():
    for i, gamepad in enumerate(input_handler.gamepads):
        gamepad_name = 'gamepad'
        if i > 0:
            gamepad_name += f'_{i}'

        # left stick
        x = gamepad.findAxis(InputDevice.Axis.left_x).value
        if abs(x) < .1:
            x = 0
        held_keys[f'{gamepad_name} left stick x'] = x

        y = gamepad.findAxis(InputDevice.Axis.left_y).value
        if abs(y) < .1:
            y = 0
        held_keys[f'{gamepad_name} left stick y'] = y

        # right stick
        x = gamepad.findAxis(InputDevice.Axis.right_x).value
        if abs(x) < .1:
            x = 0
        held_keys[f'{gamepad_name} right stick x'] = x

        y = gamepad.findAxis(InputDevice.Axis.right_y).value
        if abs(y) < .1:
            y = 0
        held_keys[f'{gamepad_name} right stick y'] = y

        held_keys[f'{gamepad_name} left trigger'] = gamepad.findAxis(InputDevice.Axis.left_trigger).value
        held_keys[f'{gamepad_name} right trigger'] = gamepad.findAxis(InputDevice.Axis.right_trigger).value

Entity(name='gamepad_handler', update=update, eternal=True) # connect update() to an entity so it runs



import time
import platform

from direct.showbase.ShowBase import ShowBase
from direct.task.Task import Task
from panda3d.core import ConfigVariableBool

from ursina.window import instance as window
from ursina import application
from ursina import input_handler
from ursina.scene import instance as scene
from ursina.camera import instance as camera
from ursina.mouse import instance as mouse
from ursina import entity


import __main__
time.dt = 0
time.dt_unscaled = 0
keyboard_keys = '1234567890qwertyuiopasdfghjklzxcvbnm'


from ursina.scripts.singleton_decorator import singleton
@singleton
class Ursina(ShowBase):
    def __init__(self, title='ursina', icon='textures/ursina.ico', borderless=True, fullscreen=False, size=None, forced_aspect_ratio=None, position=None, vsync=True, editor_ui_enabled=True, window_type='onscreen', development_mode=True, render_mode=None, show_ursina_splash=False, **kwargs):
        """The main class of Ursina. This class is a singleton, so you can only have one instance of it.

        Keyword Args (optional):
            title (str): The title of the window.\n
            fullscreen (bool): Whether the window should be fullscreen or not.\n
            size (tuple(int, int)): The size of the window.\n
            forced_aspect_ratio (bool): Whether the window should have a forced aspect ratio or not.\n
            position (tuple(int, int)): The position of the window.\n
            vsync (bool): Whether the window should have vsync enabled or not.\n
            borderless (bool): Whether the window should be borderless or not.\n
            show_ursina_splash (bool): Whether the Ursina splash should be shown or not.\n
            render_mode (str): The render mode of the window.\n
            development_mode (bool): Whether the development mode should be enabled or not.\n
            editor_ui_enabled (bool): Whether the editor UI should be enabled or not.\n
            window_type (str): The type of the window. Can be 'onscreen', 'offscreen' or 'none'.\n
        """
        entity._warn_if_ursina_not_instantiated = False
        application.window_type = window_type
        application.base = self
        application.development_mode = development_mode
        application.show_ursina_splash = show_ursina_splash

        try:
            import gltf
            gltf.patch_loader(self.loader)
        except:
            pass

        if 'gltf_no_srgb' in kwargs:
            application.gltf_no_srgb = kwargs['gltf_no_srgb']

        if not application.development_mode:
            fullscreen = True

        window.ready(title=title, icon=icon,
            borderless=borderless, fullscreen=fullscreen, size=size, forced_aspect_ratio=forced_aspect_ratio, position=position, vsync=vsync, window_type=window_type,
            editor_ui_enabled=editor_ui_enabled, render_mode=render_mode)

        super().__init__(windowType=application.window_type)
        window.apply_settings()
        # camera
        if application.window_type != 'none':
            camera._cam = self.camera
            camera._cam.reparent_to(camera)
            camera.render = self.render
            camera.position = (0, 0, -20)
            scene.camera = camera
            camera.set_up()

        # input
        if application.window_type == 'onscreen':
            self.buttonThrowers[0].node().setButtonDownEvent('buttonDown')
            self.buttonThrowers[0].node().setButtonUpEvent('buttonUp')
            self.buttonThrowers[0].node().setButtonRepeatEvent('buttonHold')
            self.buttonThrowers[0].node().setKeystrokeEvent('keystroke')
        self._input_name_changes = {
            'mouse1' : 'left mouse down', 'mouse1 up' : 'left mouse up', 'mouse2' : 'middle mouse down', 'mouse2 up' : 'middle mouse up', 'mouse3' : 'right mouse down', 'mouse3 up' : 'right mouse up',
            'wheel_up' : 'scroll up', 'wheel_down' : 'scroll down',
            'arrow_left' : 'left arrow', 'arrow_left up' : 'left arrow up', 'arrow_up' : 'up arrow', 'arrow_up up' : 'up arrow up', 'arrow_down' : 'down arrow', 'arrow_down up' : 'down arrow up', 'arrow_right' : 'right arrow', 'arrow_right up' : 'right arrow up',
            'lcontrol' : 'left control', 'rcontrol' : 'right control', 'lshift' : 'left shift', 'rshift' : 'right shift', 'lalt' : 'left alt', 'ralt' : 'right alt',
            'lcontrol up' : 'left control up', 'rcontrol up' : 'right control up', 'lshift up' : 'left shift up', 'rshift up' : 'right shift up', 'lalt up' : 'left alt up', 'ralt up' : 'right alt up',
            'control-mouse1' : 'left mouse down', 'control-mouse2' : 'middle mouse down', 'control-mouse3' : 'right mouse down',
            'shift-mouse1' : 'left mouse down', 'shift-mouse2' : 'middle mouse down', 'shift-mouse3' : 'right mouse down',
            'alt-mouse1' : 'left mouse down', 'alt-mouse2' : 'middle mouse down', 'alt-mouse3' : 'right mouse down',
            'page_down' : 'page down', 'page_down up' : 'page down up', 'page_up' : 'page up', 'page_up up' : 'page up up',
            }

        for e in keyboard_keys:
            self.accept(f'raw-{e}', self.input, [e, True])
            self.accept(f'raw-{e}-up', self.input_up, [e, True])
            self.accept(f'raw-{e}-repeat', self.input_hold, [e, True])

        self.accept('buttonDown', self.input)
        self.accept('buttonUp', self.input_up)
        self.accept('buttonHold', self.input_hold)
        self.accept('keystroke', self.text_input)
        ConfigVariableBool('paste-emit-keystrokes', False)

        self.disableMouse()
        mouse._mouse_watcher = self.mouseWatcherNode
        mouse.enabled = True
        self.mouse = mouse

        scene.set_up()
        self._update_task = self.taskMgr.add(self._update, "update")

        # try to load settings that need to be applied before entity creation
        application.load_settings()

        from ursina.prefabs.hot_reloader import HotReloader
        application.hot_reloader = HotReloader(__main__.__file__ if hasattr(__main__, '__file__') else 'None', enabled=application.development_mode)  # make sure it's running from a file and not an interactive session.

        try:
            from ursina import gamepad
        except e as Exception:
            print(e)

        if application.window_type != 'none':
            window.make_editor_gui()
            if 'use_ingame_console' in kwargs and kwargs['use_ingame_console']:
                import builtins
                from ursina import Entity, TextField, color
                window.console = Entity(parent=window.editor_ui, position=window.top_left, eternal=True)
                window.console.text_field = TextField(parent=window.console, scale=.75, font='VeraMono.ttf', max_lines=20, position=(0,0), register_mouse_input=True, text_input=None, eternal=True)
                window.console.text_field.bg.color = color.black66
                window.console.text_field.bg.scale_x = 1.5
                def _custom_print(*args, **kwargs):  # makes print() poutput to the in-game console instead of the terminal if use_ingame_console is True
                    content = ' '.join([str(e) for e in args])
                    if 'error' in content:
                        content = f'{content}'
                    window.console.text_field.cursor.y = window.console.text_field.text.count('\n')
                    window.console.text_field.cursor.x = 999
                    window.console.text_field.add_text(f'\n{content}')
                builtins.print = _custom_print

                def _console_text_input(key):
                    if key == '|':
                        window.console.text_field.enabled = not window.console.text_field.enabled
                window.console.text_input = _console_text_input


        if 'editor_ui_enabled' in kwargs:
            window.editor_ui.enabled = kwargs['editor_ui_enabled']

        print('package_folder:', application.package_folder)
        print('asset_folder:', application.asset_folder)

        entity._Ursina_instance = self


    def _update(self, task):
        """Internal task that runs every frame. Updates time, mouse, sequences and entities."""
        if application.calculate_dt:
            time.dt_unscaled = globalClock.getDt()
            time.dt = time.dt_unscaled * application.time_scale          # time between frames
        mouse.update()

        if hasattr(__main__, 'update') and __main__.update and not application.paused:
            __main__.update()

        for seq in application.sequences:
            seq.update()

        for e in scene.entities:
            if not e.enabled or e.ignore:
                continue
            if application.paused and e.ignore_paused is False:
                continue
            if e.has_disabled_ancestor():
                continue

            if hasattr(e, 'update') and callable(e.update):
                e.update()

            if hasattr(e, 'scripts'):
                for script in e.scripts:
                    if script.enabled and hasattr(script, 'update') and callable(script.update):
                        script.update()

        return Task.cont


    def input_up(self, key, is_raw=False): # internal method for key release
        if not is_raw and key in keyboard_keys:
            return

        if key in ('wheel_up', 'wheel_down'):
            return

        key += ' up'
        self.input(key)


    def input_hold(self, key, is_raw=False):   # internal method for handling repeating input that occurs when you hold the key
        key = key.replace('control-', '')
        key = key.replace('shift-', '')
        key = key.replace('alt-', '')

        if key in self._input_name_changes:
            key = self._input_name_changes[key]

        key += ' hold'
        self.input(key)


    def input(self, key, is_raw=False): # internal method for handling input
        """Built-in input handler. Propagates the input to all entities and the input function of the main script. Main use case for this it to simulate input though code, like: app.input('a').

        Args:
            key (Any): The input key.
            is_raw (bool, optional): Whether or not the input should be treated as "raw". Defaults to False.
        """
        if not is_raw and key in keyboard_keys:
            return

        if 'mouse' not in key:
            for prefix in ('control-', 'shift-', 'alt-'):
                if key.startswith(prefix):
                    key = key.replace('control-', '')
                    key = key.replace('shift-', '')
                    key = key.replace('alt-', '')
                    if key in keyboard_keys:
                        return

        if key in self._input_name_changes:
            key = self._input_name_changes[key]

        if key in input_handler.rebinds:
            key = input_handler.rebinds[key]

        input_handler.input(key)

        if not application.paused:
            if hasattr(__main__, 'input'):
                __main__.input(key)

        for e in scene.entities:
            if e.enabled is False or e.ignore or e.ignore_input:
                continue
            if application.paused and e.ignore_paused is False:
                continue
            if e.has_disabled_ancestor():
                continue


            if hasattr(e, 'input') and callable(e.input):
                if e.input(key):
                    break

            if hasattr(e, 'scripts'):
                for script in e.scripts:
                    if script.enabled and hasattr(script, 'input') and callable(script.input):
                        if script.input(key):
                            break


        mouse.input(key)


    def text_input(self, key):  # internal method for handling text input
        key_code = ord(key)
        if key_code < 32 or (key_code >= 127 and key_code <= 160):
            return

        if key == ' ' and input_handler.held_keys['control']:
            return

        if not application.paused:
            if hasattr(__main__, 'text_input'):
                __main__.text_input(key)

        for e in scene.entities:
            if e.enabled is False or e.ignore or e.ignore_input:
                continue
            if application.paused and e.ignore_paused is False:
                continue

            if hasattr(e, 'text_input') and callable(e.text_input):
                e.text_input(key)

            if hasattr(e, 'scripts'):
                for script in e.scripts:
                    if script.enabled and hasattr(script, 'text_input') and callable(script.text_input):
                        script.text_input(key)

    def step(self): # use this control the update loop yourself. call app.step() in a while loop for example, instead of app.run()
        self.taskMgr.step()


    def run(self, info=True):
        if application.show_ursina_splash:
            from ursina.prefabs import ursina_splash

        application.load_settings()
        if info:
            print('os:', platform.system())
            print('development mode:', application.development_mode)
            print('application successfully started')

        super().run()


if __name__ == '__main__':
    from ursina import *
    app = Ursina()
    def input(key):
        print(key)
    app.run()



from ursina import Entity, Draggable, Text, Slider, Button, color, Vec3, Quad, invoke
from ursina.prefabs.input_field import InputField


class Space():
    def __init__(self, height=1):
        self.height = height


class WindowPanel(Draggable):
    def __init__(self, title='', content=[], **kwargs):
        super().__init__(origin=(-0,.5), scale=(.5, Text.size*2), text=title, color=color.black)

        self.content = content
        self.popup = False
        self._prev_input_field = None
        self._original_scale = self.scale

        for key, value in kwargs.items():
            setattr(self, key ,value)

        # if self.text_entity:
        #     self.text_entity.world_scale_y = 1

        self.panel = Entity(parent=self, model='quad', origin=(0,.5), z=.1, color=self.color.tint(.1), collider='box')

        if self.popup:
            self.lock = Vec3(1,1,1)
            self.bg = Button(parent=self, z=1, scale=(999, 999), color=color.black66, highlight_color=color.black66, pressed_color=color.black66)
            self.bg.on_click = self.close

        if self.content:
            self.layout()


    def layout(self):
        content = self.content
        if not content:
            return
        spacing = .25
        height = 1 + spacing

        if isinstance(content, dict):
            content = content.values()

        for c in content:
            # print('........', c)
            if isinstance(c, Space):
                height += c.height

            if isinstance(c, Entity):
                c.world_parent = self
                c.position = (0, -height, 0)

                if isinstance(c, InputField):
                    if self._prev_input_field:
                        self._prev_input_field.next_field = c
                    self._prev_input_field = c

                if isinstance(c, Text):
                    c.origin = (-.5, .5)
                    c.x = -.48
                    height += len(c.lines)

                elif isinstance(c, Button):
                    c.world_parent = self
                    c.scale = (.9, 1)
                    if hasattr(c, 'height'):
                        c.scale_y = height
                    c.model = Quad(aspect=c.world_scale_x/c.world_scale_y)
                    height += c.scale_y
                    # c.y -= c.scale_y/2

                elif isinstance(c, Slider):
                    c.world_parent = self
                    c.x = -.5 * .9
                    c.scale = (.9*2, 20)
                    # print('-------------', c.scale_y * c.height)
                    height += 1

                elif hasattr(c, 'scale_y'):
                    height += c.scale_y

                if hasattr(c, 'text_entity') and c.text_entity is not None:
                    c.text_entity.world_scale = Vec3(20,20,1)

                height += spacing

        self.panel.scale_y = height
        self.panel.model = Quad(aspect=self.panel.world_scale_x/self.panel.world_scale_y, radius=.025)
        self.panel.origin = (0, .5)


    def on_enable(self):
        pass
        try:
            if self.popup:
                self.bg.enabled = True
                # self.animate_scale(self._original_scale, duration=.1)
        except:
            pass

    def close(self):
        if self.popup:
            self.bg.enabled = False
        # self.animate_scale_y(0, duration=.1)
        invoke(setattr, self, 'enabled', False, delay=.2)


if __name__ == '__main__':
    '''
    WindowPanel is an easy way to create UI. It will automatically layout the content.
    '''
    from ursina import Ursina, ButtonGroup
    app = Ursina()
    wp = WindowPanel(
        title='Custom Window',
        content=(
            Text('Name:'),
            InputField(name='name_field'),
            Button(text='Submit', color=color.azure),
            Slider(),
            Slider(),
            ButtonGroup(('test', 'eslk', 'skffk'))
            ),
        )
    wp.y = wp.panel.scale_y / 2 * wp.scale_y    # center the window panel

    app.run()



import operator
from math import sqrt, sin, acos, pi, cos, floor
from panda3d.core import Vec4, LVector3f, Mat3, Mat4
from ursina.vec2 import Vec2
from ursina.vec3 import Vec3
from ursina import color
from ursina.color import Color
internal_sum = sum


def distance(a, b):
    if isinstance(a, (Color, Vec4)):
        dist = abs(a[0] - b[0])
        dist += abs(a[1] - b[1])
        dist += abs(a[2] - b[2])
        dist += abs(a[3] - b[3])
        return dist

    # if input is Entity, convert to positions
    if hasattr(a, 'world_position'): a = a.world_position
    if hasattr(b, 'world_position'): b = b.world_position

    dist = sqrt((b[0] - a[0])**2 + (b[1] - a[1])**2 + (b[2] - a[2])**2)
    # print('------------DIST:', dist)
    return dist


def distance_2d(a, b):
    if hasattr(a, 'position'): a = a.position
    if hasattr(b, 'position'): b = b.position

    return sqrt((b[0] - a[0])**2 + (b[1] - a[1])**2)


def distance_xz(a, b):
    if hasattr(a, 'position'): a = a.position
    if hasattr(b, 'position'): b = b.position

    return sqrt((b[0] - a[0])**2 + (b[2] - a[2])**2)


def lerp(a, b, t):
    if isinstance(a, (int, float, complex)):
        return a + (b - a) * t

    elif isinstance(a, Color) and isinstance(b, Color):
        col = [lerp(e[0], e[1], t) for e in zip(a,b)]
        return Color(col[0], col[1], col[2], col[3])

    elif isinstance(a, (tuple, list, Vec2, Vec3, Vec4, LVector3f)) and isinstance(b, (tuple, list, Vec2, Vec3, Vec4, LVector3f)):
        lerped = []
        for i in range(min(len(a), len(b))):
            lerped.append(lerp(a[i], b[i], t))

        if isinstance(a, (tuple, list)):
            return type(a)(lerped)
        else:
            return type(a)(*lerped)
    else:
        print(f'''can't lerp types {type(a)} and {type(b)}''')


def inverselerp(a, b, t) :
    return (a - b) / (t - b)


def slerp(q1, q2, t):
    costheta = q1.dot(q2)
    if costheta < 0.0:
        costheta = -costheta
        q1 = q1.conjugate()
    elif costheta > 1.0:
        costheta = 1.0

    theta = acos(costheta)
    if abs(theta) < 0.01:
        return q2

    sintheta = sqrt(1.0 - costheta * costheta)
    if abs(sintheta) < 0.01:
        return (q1+q2)*0.5

    r1 = sin((1.0 - t) * theta) / sintheta
    r2 = sin(t * theta) / sintheta
    return (q1*r1) + (q2*r2)



def clamp(value, floor, ceiling):
    return max(min(value, ceiling), floor)


def round_to_closest(value, step=0):
    if not step:
        return value

    step = 1/step
    return round(value * step) / step


def rotate_around_point_2d(point, origin, deg):
    angle_rad = -deg/180 * pi # ursina rotation is positive=clockwise, so do *= -1
    cos_angle = cos(angle_rad)
    sin_angle = sin(angle_rad)
    dx = point[0] - origin[0]
    dy = point[1] - origin[1]

    return (
        origin[0] + (dx*cos_angle - dy*sin_angle),
        origin[1] + (dx*sin_angle + dy*cos_angle)
        )

def world_position_to_screen_position(point): # get screen position(ui space) from world space.
    from ursina import camera, Entity, destroy
    _temp_entity = Entity(position=point, add_to_scene_entities=False)
    result = _temp_entity.screen_position
    destroy(_temp_entity)
    return result


def sum(l):
    try:
        return internal_sum(l)
    except:
        pass

    total = l[0].__class__()
    for e in l:
        total += e

    return total


def make_gradient(index_color_dict):    # returns a list of 256 colors
    '''
    given a dict of positions and colors, interpolates the colors into a list of 256 colors
    example input: {'0':color.hex('#9d9867'), '38':color.hex('#828131'), '54':color.hex('#5d5b2a'), '255':color.hex('#000000')}
    '''
    gradient = [color.black for i in range(256)]

    gradient_color_keys = tuple(index_color_dict.keys())
    gradient_color_values = tuple(index_color_dict.values())

    for i in range(len(gradient_color_values)):
        from_col = color.hex(gradient_color_values[i-1])
        to_col = color.hex(gradient_color_values[i])

        from_num = 0
        if i > 0:
            from_num = int(gradient_color_keys[i-1])
        to_num = int(gradient_color_keys[i])
        dist = to_num - from_num

        for j in range(dist):
            gradient[from_num+j] = lerp(from_col, to_col, j/dist)

    return gradient


def sample_gradient(list_of_values, t):     # distribute list_of_values equally on a line and get the interpolated value at t (0-1).
    l = len(list_of_values)
    if l == 1:
        return list_of_values[0]

    t *= l-1
    index = floor(t - .001)
    index = clamp(index, 0, l-1)
    relative = t - index
    # print(t, index, relative)

    if index < l-1:
        return lerp(list_of_values[index], list_of_values[index+1], relative)
    else:
        return lerp(list_of_values[index-1], list_of_values[index], relative)





class Bounds:
    __slots__ = ['start', 'end', 'center', 'size']
    def __init__(self, start, end, center, size):
        self.start = start
        self.end = end
        self.center = center
        self.size = size


if __name__ == '__main__':
    from ursina import *
    app = Ursina()
    e1 = Entity(position = (0,0,0))
    e2 = Entity(position = (0,1,1))
    distance(e1, e2)
    distance_xz(e1, e2.position)

    between_color = lerp(color.lime, color.magenta, .5)
    print(between_color)
    print(lerp((0,0), (0,1), .5))
    print(lerp(Vec2(0,0), Vec2(0,1), .5))
    print(lerp([0,0], [0,1], .5))

    print(round(Vec3(.38, .1351, 353.26), 2))

    p = (1,0)
    print(p, 'rotated ->', rotate_around_point_2d(p, (0,0), 90))

    app.run()



from ursina import Entity, Sequence, Func, Wait


class SpriteSheetAnimation(Entity):
    def __init__(self, texture, animation_clips, tileset_size=[4,1], fps=12, model='quad', autoplay=True, **kwargs):
        kwargs['model'] = model
        kwargs['texture'] = texture
        kwargs['tileset_size'] = tileset_size
        super().__init__(**kwargs)

        self.animation_clips = animation_clips # should be a dict
        self.current_animation = None

        for key, value in self.animation_clips.items():
            start_coord, end_coord = value
            s = Sequence(loop=True)
            s.name = key
            # print(s.name)

            for y in range(start_coord[1], end_coord[1]+1):
                for x in range(start_coord[0], end_coord[0]+1):
                    s.extend([
                        Func(setattr, self, 'tile_coordinate', (x,y)),
                        Wait(1/fps)
                    ])
            self.animation_clips[key] = s
            self.animations.append(s)
        print('-------------', self.animation_clips)


    def play_animation(self, animation_name):
        if not self.animation_clips:
            return

        [anim.pause() for anim in self.animation_clips.values()]
        self.current_animation = self.animation_clips[animation_name]

        # if restart:
        self.animation_clips[animation_name].start()
        # else:
        #     self.animation_clips[animation_name].resume()


    @property
    def state(self):
        return self.current_animation

    @state.setter
    def state(self, value):
        if not self.current_animation:
            self.play_animation(value)
            return

        if value != self.current_animation.name:
            print('set state to:', value, self.current_animation.name)
            self.play_animation(value)




if __name__ == '__main__':
    '''
    Sprite sheet coordinate system:
    (0,3) (1,3) (2,3) (3,3)
    (0,2) (1,2) (2,2) (3,2)
    (0,1) (1,1) (2,1) (3,1)
    (0,0) (1,0) (2,0) (3,0)
    '''
    from ursina import Ursina, held_keys, Vec2, time
    app = Ursina()
    class Player(Entity):
        def __init__(self):
            super().__init__()
            self.direction = Vec2(0,0)
            self.velocity = Vec2(0,0)

            self.model = 'quad'
            self.texture='arrow_right'

            self.graphics = SpriteSheetAnimation('example_spritesheet', tileset_size=(4,4), fps=6, animation_clips={
                'idle_up' : ((1,0), (1,0)),     # makes an animation from (0,0) to (0,0), a single frame
                'idle_right' : ((1,1), (1,1)),  # makes an animation from (0,0) to (0,0), a single frame
                'idle_left' : ((1,2), (1,2)),   # makes an animation from (0,0) to (0,0), a single frame
                'idle_down' : ((1,3), (1,3)),   # makes an animation from (0,0) to (0,0), a single frame
                'walk_up' : ((0,0), (3,0)),     # makes an animation from (0,0) to (3,0), the bottom row
                'walk_right' : ((0,1), (3,1)),
                'walk_left' : ((0,2), (3,2)),
                'walk_down' : ((0,3), (3,3)),
            },
            parent=self,
            )

        def input(self, key):
            if key == 'd':      self.velocity = Vec2(1,0)
            # elif key == 'd up':   self.velocity[0] = -held_keys['a']
            elif key == 'a':      self.velocity = Vec2(-1,0)
            # if key == 'a up':   self.velocity[0] = held_keys['d']
            elif key == 'w':      self.velocity = Vec2(0,1)
            if key == 'w up':   self.velocity = Vec2()
            elif key == 's':      self.velocity[1] = -1
            # if key == 's up':   self.velocity[1] = held_keys['w']

            # if held_keys['d'] or held_keys['a'] or held_keys['w'] or held_keys['s']:

            if key in 'wasd':
                self.prev_key = key

            elif key in ('d up', 'a up', 'w up', 's up'):
                self.velocity = Vec2(held_keys['d']-held_keys['a'], held_keys['w']-held_keys['s'])

            # if held_keys['d'] or held_keys['a']:    self.scale_x = self.velocity.x
            # if held_keys['w'] or held_keys['s']:    self.scale_y = self.velocity.y
            # if key in 'wasd':
            #     self.direction = {
            #         'w' : Vec2(0,1),
            #         'a' : Vec2(-1,0),
            #         'd' : Vec2(1,0),
            #         's' : Vec2(0,-1),
            #     }[key]
            # else:
            #     self.direction = Vec2(0,0)
            # if key == 'd':
            #     self.velocity = Vec2(1,0)
            #     # self.scale_x = self._original_scale_x
            # if key == 'd up':
            #     self.velocity = Vec2(-held_keys['a'], 0)
            #
            # if key == 'a':
            #     self.velocity = Vec2(-1, 0)
            #     # self.scale_x = -self._original_scale_x
            # if key == 'a up':
            #     self.velocity = Vec2(held_keys['d'], held_keys['s'])

            # if held_keys['d'] or held_keys['a']:
            #     self.scale_x = self._original_scale_x * self.velocity

        def update(self):
            # go to a specific animation while holding a key.
            # setting .state will not restart the animation like .play_animation would.
            # while this could be done with .play_animation in input, this way we can have
            # the animation code at the same place as the movement code, which would likely be in update.
            self.position += self.velocity * time.dt * 5
            # direction = ''.join([held_keys['w'],held_keys['a'],held_keys['s'], held_keys['d']])
            if held_keys['w']:
                self.graphics.state = 'walk_up'
            elif held_keys['s']:
                self.graphics.state = 'walk_down'
            elif held_keys['d']:
                self.graphics.state = 'walk_right'
            elif held_keys['a']:
                self.graphics.state = 'walk_left'

            # if not moving, go to an idle animation depending on which direction we're facing
            elif self.graphics.current_animation:
                if self.graphics.state.name == 'walk_up':
                    self.graphics.state = 'idle_up'
                elif self.graphics.state.name == 'walk_right':
                    self.graphics.state = 'idle_right'
                elif self.graphics.state.name == 'walk_left':
                    self.graphics.state = 'idle_left'
                elif self.graphics.state.name == 'walk_down':
                    self.graphics.state = 'idle_down'


        # elif key == 'w up':
        #     self.graphics.play_animation('idle')

    # print(self.graphics.animations['walk_up'].funcs)
    player = Player()
    e = Entity(model='quad', texture='example_spritesheet', x=-1)

    # print(self.graphics.animations)
    # from ursina import destroy
    # destroy(self.graphics)
    app.run()



from ursina.entity import Entity
from ursina.scene import instance as scene
from ursina.vec3 import Vec3
from ursina import color
from ursina.ursinastuff import invoke


_boxcast_box = Entity(model='cube', origin_z=-.5, collider='box', color=color.white33, enabled=False, eternal=True, add_to_scene_entities=False)

def boxcast(origin, direction=(0,0,1), distance=9999, thickness=(1,1), traverse_target=scene, ignore:list=None, debug=False): # similar to raycast, but with width and height
    if not ignore:
        ignore = []

    if isinstance(thickness, (int, float, complex)):
        thickness = (thickness, thickness)

    _boxcast_box.enabled = True
    _boxcast_box.collision = True
    _boxcast_box.position = origin
    _boxcast_box.scale = Vec3(abs(thickness[0]), abs(thickness[1]), abs(distance))
    _boxcast_box.always_on_top = debug
    _boxcast_box.visible = debug

    _boxcast_box.look_at(origin + direction)
    hit_info = _boxcast_box.intersects(traverse_target=traverse_target, ignore=ignore)

    if debug:
        _boxcast_box.collision = False
        invoke(setattr, _boxcast_box, 'enabled', False, delay=.2)
    else:
        _boxcast_box.enabled = False

    return hit_info



if __name__ == '__main__':
    from ursina import Ursina, held_keys, camera, duplicate, raycast, time, EditorCamera
    app = Ursina()

    '''
    Casts a ray from *origin*, in *direction*, with length *distance* and returns
    a HitInfo containing information about what it hit. This ray will only hit entities with a collider.

    Use optional *traverse_target* to only be able to hit a specific entity and its children/descendants.
    Use optional *ignore* list to ignore certain entities.
    Setting debug to True will draw the line on screen.

    Example where we only move if a wall is not hit:
    '''


    class Player(Entity):

        def update(self):
            self.direction = Vec3(
                self.forward * (held_keys['w'] - held_keys['s'])
                + self.right * (held_keys['d'] - held_keys['a'])
                ).normalized()  # get the direction we're trying to walk in.

            origin = self.world_position + (self.up*.5) # the ray should start slightly up from the ground so we can walk up slopes or walk over small objects.
            hit_info = raycast(origin , self.direction, ignore=(self,), distance=.5, debug=False)
            if not hit_info.hit:
                self.position += self.direction * 5 * time.dt

    Player(model='cube', origin_y=-.5, color=color.orange)
    wall_left = Entity(model='cube', collider='box', scale_y=3, origin_y=-.5, color=color.azure, x=-4)
    wall_right = duplicate(wall_left, x=4)
    camera.y = 2
    app.run()

    # test
    breakpoint()
    d = Entity(parent=scene, position=(0,0,2), model='cube', color=color.orange, collider='box', scale=2)
    e = Entity(model='cube', color=color.lime)

    camera.position = (0, 15, -15)
    camera.look_at(e)
    # camera.reparent_to(e)
    speed = .01
    rotation_speed = 1
    intersection_marker = Entity(model='cube', scale=.2, color=color.red)

    def update():
        e.position += e.forward * held_keys['w'] * speed
        e.position += e.left * held_keys['a'] * speed
        e.position += e.back * held_keys['s'] * speed
        e.position += e.right * held_keys['d'] * speed

        e.rotation_y -= held_keys['q'] * rotation_speed
        e.rotation_y += held_keys['e'] * rotation_speed

        # ray = raycast(e.world_position, e.forward, 3, debug=True)
        # ray = raycast(e.world_position, e.forward, 3, debug=True)
        ray = boxcast(e.world_position, e.right, 3, debug=True)
        # print(ray.distance, ray2.distance)
        intersection_marker.world_position = ray.world_point
        intersection_marker.visible = ray.hit
        if ray.hit:
            d.color = color.azure
        else:
            d.color = color.orange

    t = time.time()
    # ray = raycast(e.world_position, e.forward, 3, debug=True)
    print(time.time() - t)
    # raycast((0,0,-2), (0,0,1), 5, debug=False)

    EditorCamera()
    app.run()



from ursina import *
from time import perf_counter

class TrailRenderer(Entity):
    def __init__(self, thickness=10, color=color.white, end_color=color.clear, length=6, **kwargs):
        super().__init__(**kwargs)
        self.renderer = Entity(
            model = Mesh(
            vertices=[self.world_position for i in range(length)],
            colors=[lerp(end_color, color, i/length*2) for i in range(length)],
            mode='line',
            thickness=thickness,
            static=False
            )
        )
        self._t = 0
        self.update_step = .025


    def update(self):
        self._t += time.dt
        if self._t >= self.update_step:
            t = perf_counter()
            self._t = 0
            self.renderer.model.vertices.pop(0)
            self.renderer.model.vertices.append(self.world_position)
            self.renderer.model.generate()
            print('---', perf_counter() - t)

    def on_destroy(self):
        destroy(self.renderer)



if __name__ == '__main__':
    app = Ursina()
    window.color = color.black
    mouse.visible = False
    player = Entity()
    player.graphics = Entity(parent=player, scale=.1, model='circle')
    trail_renderer = TrailRenderer(parent=player, thickness=100, color=color.yellow, length=6)

    pivot = Entity(parent=player)
    trail_renderer = TrailRenderer(parent=pivot, x=.1, thickness=20, color=color.orange)
    trail_renderer = TrailRenderer(parent=pivot, y=1, thickness=20, color=color.orange)
    trail_renderer = TrailRenderer(parent=pivot, thickness=2, color=color.orange, alpha=.5, position=(.4,.8))
    trail_renderer = TrailRenderer(parent=pivot, thickness=2, color=color.orange, alpha=.5, position=(-.5,.7))

    def update():
        player.position = lerp(player.position, mouse.position*10, time.dt*4)

        if pivot:
            pivot.rotation_z -= 3
            pivot.rotation_x -= 2

    def input(key):
        if key == 'space':
            destroy(pivot)


    app.run()



class Scene(Entity):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        Entity(parent=self, position=Vec3(-1.0334,0,3.0372), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-5.3896,0.002,9.6256), scale=Vec3(1,2,1), model='cube', origin=Vec3(0,-0.5,0), color=color.hsv(30,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-2.1674,0.064,2.3704), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(0.3745,0.002,-6.6139), model='cube', color=color.hsv(209,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-0.2384,0.065,-0.8437), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(1.1839,0.003,8.2525), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(0.0499,0.067,7.5857), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(4.3956,0.004,-2.4644), model='cube', color=color.hsv(209,1,1), ignore=True)
        Entity(parent=self, position=Vec3(2.1659,0.067,-2.8756), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-1.4502,0.069,-6.4631), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-6.3671,0.001,4.0369), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-3.4524,0.001,3.39), scale=Vec3(1,2,1), model='cube', color=color.hsv(30,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-7.5011,0.065,3.3701), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-4.7031,0.002,0.8228), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-5.8371,0.066,0.156), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-2.5641,0.003,2.7849), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-3.6981,0.067,2.1181), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-2.2987,0.004,-1.2091), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-3.4327,0.068,-1.8759), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-4.8313,0.005,-2.0694), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-5.9653,0.069,-2.7362), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-1.9876,-0.067,13.2403), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(0.9271,-0.067,12.5934), scale=Vec3(1,2,1), model='cube', color=color.hsv(30,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-1.5359,-0.004,6.1062), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(-0.3235,-0.066,10.0263), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(0.1282,-0.003,2.8922), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(1.8154,-0.065,11.9884), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(0.6814,-0.001,11.3216), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(2.0808,-0.064,7.9944), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(0.9468,0,7.3276), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(1.3989,-0.064,0.6668), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(3.1222,0.002,2.838), scale=Vec3(1,2,1), model='cube', color=color.hsv(30,1,1), ignore=True)
        Entity(parent=self, position=Vec3(4.0105,0.004,2.2329), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(2.8765,0.068,1.5661), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(0.839,-0.066,7.2207), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(3.7537,-0.066,6.5738), scale=Vec3(1,2,1), model='cube', color=color.hsv(30,1,1), ignore=True)
        Entity(parent=self, position=Vec3(2.5031,-0.065,4.0067), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(4.642,-0.064,5.9688), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(3.508,0,5.302), model='cube', color=color.hsv(330,1,1), ignore=True)
        Entity(parent=self, position=Vec3(4.9074,-0.063,1.9748), model='cube', color=color.hsv(0,1,1), ignore=True)
        Entity(parent=self, position=Vec3(3.7734,0.001,1.308), model='cube', color=color.hsv(330,1,1), ignore=True)



from ursina import *
import pyperclip
from PIL import Image
from copy import deepcopy
import sys
from math import floor
from ursina.shaders import unlit_shader
import json


from ursina.scripts.property_generator import generate_properties_for_class
@generate_properties_for_class()
class GridEditor(Entity):
    def __init__(self, size=(32,32), palette=(' ', '#', '|', 'o'), canvas_color=color.white, **kwargs):
        super().__init__(parent=camera.ui, position=(-.45,-.45), scale=.9, model='quad', origin=(-.5,-.5), visible_self=False)
        self.w, self.h = int(size[0]), int(size[1])
        self.canvas = Entity(parent=self, model='quad', origin=(-.5,-.5), shader=unlit_shader, scale=(self.w/self.h, 1), color=canvas_color)
        self.canvas_collider = Entity(parent=self.canvas, model='wireframe_quad', origin=self.canvas.origin, color=color.blue, scale=2, position=(-.5,-.5), collider='box')
        sys.setrecursionlimit(max(sys.getrecursionlimit(), self.w * self.h))
        # self.grid = [[palette[0] for x in range(self.w)] for y in range(self.h)]
        if not hasattr(self, 'grid'):
            self.grid = [[palette[0] for y in range(self.h)] for x in range(self.w)]
        self.brush_size = 1
        self.auto_render = True
        self.cursor = Entity(parent=self.canvas, model=Quad(segments=0, mode='line', thickness=2), origin=(-.5,-.5), scale=(1/self.w, 1/self.h), color=color.color(120,1,1,.5), z=-.2, shader=unlit_shader)

        self.selected_char = palette[1]
        self.palette = palette
        self.start_pos = None
        self.prev_draw = None
        self.lock_axis = None
        self.outline = Entity(parent=self.canvas, model=Quad(segments=0, mode='line', thickness=2), color=color.cyan, z=.01, origin=(-.5,-.5))

        self.rect_selection = [Vec2(0,0), Vec2(0,0)]
        self.selection_renderer = Entity(parent=self.canvas, model=Mesh(mode='line', thickness=2), color=color.lime, alpha=.5, z=-.01, origin=(-.5,-.5), scale=(1/self.w,1/self.h))
        self.rect_tool = Entity(parent=self.canvas, model=Quad(0, mode='line', thickness=2), color=color.lime, z=-.01, origin=(-.5,-.5), start=Vec2(0,0), end=Vec2(0,0))
        # self.selection_mover = Draggable(parent=self.canvas, model='circle', color=color.blue, origin=(.5,.5), step=(1/self.w,1/self.h,0), enabled=False)
        self.selection_matrix = [[0 for y in range(self.h)] for x in range(self.w)]
        self.temp_paste_layer = Entity(parent=self.cursor, model='quad', origin=(-.5,-.5), z=-.02, enabled=1)
        Entity(parent=self.temp_paste_layer, model='wireframe_quad', origin=self.temp_paste_layer.origin, color=color.black)
        self.is_in_paste_mode = False

        self.undo_stack = []
        self.undo_stack.append(deepcopy(self.grid))
        self.undo_index = 0

        self.help_icon = Button(parent=self.canvas, scale=.025, model='circle', origin=(-.5,-.5), position=(-.0,1.005,-1), text='?', target_scale=.025)
        self.help_icon.tooltip = Tooltip(
            text=dedent('''
                left mouse:    draw
                control(hold): draw lines
                alt(hold):     select character
                right click:   select character
                ctrl + z:      undo
                ctrl + y:      redo
            '''),
            font='VeraMono.ttf',
            wordwrap=100,
            # scale=.75,
            )

        self.edit_mode = True

        for key, value in kwargs.items():
            setattr(self, key ,value)


    def palette_setter(self, value):
        self._palette = value
        if hasattr(self, 'palette_parent'):
            destroy(self.palette_parent)

        self.palette_parent = Entity(parent=self, position=(-.3,.5,-1), shader=unlit_shader)
        for i, e in enumerate(value):
            if isinstance(e, str):
                i = e

            b = Button(parent=self.palette_parent, scale=.05, text=i, model='quad', color=color._32, shader=unlit_shader)
            b.on_click = Func(setattr, self, 'selected_char', e)
            b.tooltip = Tooltip(str(e))

            if isinstance(e, Color):
                b.color = e

        grid_layout(self.palette_parent.children, max_x=4)

    def edit_mode_getter(self):
        return getattr(self, '_edit_mode', True)

    def edit_mode_setter(self, value):
        self._edit_mode = value
        self.cursor.enabled = value
        self.outline.enabled = value
        self.palette_parent.enabled = value


    def update(self):
        if not self.edit_mode:
            return

        self.cursor.enabled = self.canvas_collider.hovered
        if self.canvas_collider.hovered:
            self.cursor.position = mouse.point*2 - Vec3(.5,.5,0)

            # center cursor
            self.cursor.x -= self.brush_size // 2 / self.w
            self.cursor.y -= self.brush_size // 2 / self.h

            self.cursor.x = floor(self.cursor.x * self.w) / self.w
            self.cursor.y = floor(self.cursor.y * self.h) / self.h


            if self.start_pos and mouse.left:
                if held_keys['shift'] and self.prev_draw:
                    if not self.lock_axis:
                        if abs(mouse.velocity[0]) > abs(mouse.velocity[1]):
                            self.lock_axis = 'horizontal'
                        else:
                            self.lock_axis = 'vertical'

                    if self.lock_axis == 'horizontal':
                        self.cursor.y = self.prev_draw[1] / self.h

                    if self.lock_axis == 'vertical':
                        self.cursor.x = self.prev_draw[0] / self.w


                y = int(round(self.cursor.y * self.h))
                x = int(round(self.cursor.x * self.w))
                if x < 0 or x > self.w-1 or y < 0 or y > self.h-1:
                    return

                if not held_keys['alt'] and not self.is_in_paste_mode:
                    if self.prev_draw is not None and distance_2d(self.prev_draw, (x,y)) > 1:
                        dist = distance_2d(self.prev_draw, (x,y))

                        if dist > 1: # draw line
                            for i in range(int(dist)+1):
                                inbetween_pos = lerp(self.prev_draw, (x,y), i/dist)
                                self.draw(int(inbetween_pos[0]), int(inbetween_pos[1]))

                            self.draw(x, y)
                            self.prev_draw = (x,y)

                    else:
                        self.draw(x, y)
                        self.prev_draw = (x,y)

                else:   # sample color
                    self.selected_char = self.grid[x][y]

        if mouse.right:     # selection
            self.rect_selection[1] = self.get_cursor_position()
            if self.rect_selection[0] == self.rect_selection[1]:
                self.rect_tool.start = Vec2(0,0)
                self.rect_tool.end = Vec2(0,0)
                self.rect_tool.enabled = False
                return

            self.rect_tool.start = Vec2(min(self.rect_selection[0].x, self.rect_selection[1].x), min(self.rect_selection[0].y, self.rect_selection[1].y))
            self.rect_tool.end =   Vec2(max(self.rect_selection[0].x, self.rect_selection[1].x)+1, max(self.rect_selection[0].y, self.rect_selection[1].y)+1)
            self.rect_tool.start.x = clamp(self.rect_tool.start.x, 0, self.w)
            self.rect_tool.start.y = clamp(self.rect_tool.start.y, 0, self.h)
            self.rect_tool.end.x = clamp(self.rect_tool.end.x, 0, self.w)
            self.rect_tool.end.y = clamp(self.rect_tool.end.y, 0, self.h)
            self.rect_tool.enabled = True
            self.rect_tool.position = self.rect_tool.start / Vec2(self.w, self.h)
            self.rect_tool.scale = (self.rect_tool.end - self.rect_tool.start) / Vec2(self.w, self.h)


    def get_cursor_position(self):
        y = int(round(self.cursor.y * self.h))
        x = int(round(self.cursor.x * self.w))
        return Vec2(x,y)


    def draw(self, x, y):
        for _y in range(y, min(y+self.brush_size, self.h)):
            for _x in range(x, min(x+self.brush_size, self.w)):
                self.grid[_x][_y] = self.selected_char

        if self.auto_render:
            self.render()


    def input(self, key):
        if key == 'tab':
            self.edit_mode = not self.edit_mode

        if not self.edit_mode:
            return

        if key == 'left mouse down' and not self.is_in_paste_mode:
            self.start_pos = self.get_cursor_position()
            if not held_keys['control']:
                self.prev_draw = None

        if key == 'left mouse up' and self.start_pos:
            self.start_pos = None
            self.lock_axis = None
            self.render()

            if not held_keys['control']:
                self.record_undo()

        if key == 'right mouse down' and not self.is_in_paste_mode:
            if not held_keys['shift'] and not held_keys['alt']:
                self.selection_mode = 'overwrite'
                self.clear_selection()
            elif (held_keys['control'] or held_keys['shift']) and not held_keys['alt']:
                self.selection_mode = 'add'
            elif held_keys['alt'] and not (held_keys['control'] or held_keys['shift']):
                self.selection_mode = 'subtract'

            self.rect_selection[0] = self.get_cursor_position()


        if key == 'left mouse down' and self.is_in_paste_mode:
            self.exit_paste_mode()


        if key == 'right mouse down' and self.is_in_paste_mode:
            self.exit_paste_mode(discard=True)


        elif key == 'right mouse up':
            self.rect_tool.enabled = False
            # if self.selection_mode == 'overwrite':
            #     self.clear_selection()

            new_value = int(self.selection_mode == 'overwrite' or self.selection_mode == 'add')
            for y in range(self.rect_tool.start.Y, self.rect_tool.end.Y):
                for x in range(self.rect_tool.start.X, self.rect_tool.end.X):
                    self.selection_matrix[x][y] = new_value

            self.render_selection()


        if key == 'shift up':
            self._lock_origin = None

        if held_keys['control'] and key == 'z':
            self.undo_index -= 1
            self.undo_index = clamp(self.undo_index, 0, len(self.undo_stack)-1)
            self.grid = deepcopy(self.undo_stack[self.undo_index])
            self.render()

        if held_keys['control'] and key == 'y':
            self.undo_index += 1
            self.undo_index = clamp(self.undo_index, 0, len(self.undo_stack)-1)
            self.grid = deepcopy(self.undo_stack[self.undo_index])
            self.render()

        # fill
        if key == 'g' and self.canvas_collider.hovered:
            y = int(self.cursor.y * self.h)
            x = int(self.cursor.x * self.w)
            self.floodfill(self.grid, x, y)
            self.render()
            self.record_undo()

        if key == 'x' and self.brush_size > 1:
            self.brush_size -= 1
            self.cursor.scale = Vec2(self.brush_size / self.w, self.brush_size / self.h)
            self.prev_draw = None

        if key == 'd' and self.brush_size <  8:
            self.brush_size += 1
            self.cursor.scale = Vec2(self.brush_size / self.w, self.brush_size / self.h)
            self.prev_draw = None

        if held_keys['control'] and key == 's':
            if hasattr(self, 'save'):
                self.save()

        if held_keys['control'] and key == 'c':
            self.copy()

        if held_keys['control'] and key == 'v':
            self.enter_paste_mode()


    def record_undo(self):
        self.undo_index += 1
        self.undo_stack = self.undo_stack[:self.undo_index]
        self.undo_stack.append(deepcopy(self.grid))



    def floodfill(self, matrix, x, y, first=True):
        if matrix[x][y] == self.selected_char:
            return

        if first:
            self.fill_target = matrix[x][y]

        if matrix[x][y] == self.fill_target:
            matrix[x][y] = self.selected_char
            # recursively invoke flood fill on all surrounding cells
            if x > 0:
                self.floodfill(matrix, x-1, y, first=False)
            if x < self.w-1:
                self.floodfill(matrix, x+1, y, first=False)
            if y > 0:
                self.floodfill(matrix, x, y-1, first=False)
            if y < self.h-1:
                self.floodfill(matrix, x, y+1, first=False)

    def copy(self):
        # crop matrix
        # find the indices of the rows and columns with True values
        rows = [i for i in range(len(self.selection_matrix)) if any(self.selection_matrix[i])]
        cols = [j for j in range(len(self.selection_matrix[0])) if any(self.selection_matrix[i][j] for i in range(len(self.selection_matrix)))]

        # crop the matrix based on the boolean values
        start_x = min(rows)
        start_y = min(cols)
        # end_x = max(rows) + 1
        # end_y = max(cols) + 1
        selection_width =  (max(rows) + 1) - start_x
        selection_height = (max(cols) + 1) - start_y
        # cropped_matrix = [[matrix[i][j] for j in range(start_y, max(cols) + 1)] for i in range(start_x, max(rows) + 1)]

        copy_data = [[tuple(self.grid[start_x+x][start_y+y]) if self.selection_matrix[x][y] else None for y in range(selection_height)] for x in range(selection_width)]
        pyperclip.copy(json.dumps(dict(data=copy_data)))


    def enter_paste_mode(self):
        self.is_in_paste_mode = True
        self.paste_data = json.loads(pyperclip.paste())
        if not self.paste_data['data']:
            return
        self.paste_data = self.paste_data['data']
        w = len(self.paste_data)
        h = len(self.paste_data[1])

        self.temp_paste_layer.enabled = True
        self.temp_paste_layer.scale = (w,h)

        if not self.temp_paste_layer.texture or self.temp_paste_layer.texture.size != (w,h):
            self.temp_paste_layer.texture = Texture(Image.new(mode='RGBA', size=(w,h), color=(0,0,0,0)))

        for x in range(w):
            for y in range(h):
                pixel = self.paste_data[x][y]
                if not pixel:
                    pixel = color.clear

                self.temp_paste_layer.texture.set_pixel(x, y, pixel)
        self.temp_paste_layer.texture.apply()


    def exit_paste_mode(self, discard=False):
        self.is_in_paste_mode = False
        self.temp_paste_layer.enabled = False
        x_offset = int(self.cursor.x * self.w)
        y_offset = int(self.cursor.y * self.h)
        w = len(self.paste_data)
        h = len(self.paste_data[1])

        if discard:
            return

        for x in range(w):
            for y in range(h):
                real_x = x + x_offset
                real_y = y + y_offset

                if real_x < 0 or real_x >= self.w or real_y < 0 or real_y >= self.h:
                    continue

                pixel = self.paste_data[x][y]
                if not pixel:
                    continue

                self.grid[x+x_offset][y+y_offset] = pixel

        self.render()


    def clear_selection(self):
        self.selection_matrix = [[0 for y in range(self.h)] for x in range(self.w)]
        self.selection_renderer.model.clear()
        print('cleare selection')
        # print('seofijsoefisjeofij', self.selection_renderer.enabled)


    def render_selection(self):
        quad = load_model('quad', use_deepcopy=True)
        # self.selection_renderer.enabled = True
        self.selection_renderer.model.clear(False)
        verts = []
        for x in range(self.w):
            for y in range(self.h):
                if self.selection_matrix[x][y]:
                    if x <= 0 or not self.selection_matrix[x-1][y]:
                        verts.extend((Vec3(x,y,0), Vec3(x,y+1,0)))
                    if x >= self.w-1 or not self.selection_matrix[x+1][y]:
                        verts.extend((Vec3(x+1,y,0), Vec3(x+1,y+1,0)))
                    if y <= 0 or not self.selection_matrix[x][y-1]:
                        verts.extend((Vec3(x,y,0), Vec3(x+1,y,0)))
                    if y >= self.h-1 or not self.selection_matrix[x][y+1]:
                        verts.extend((Vec3(x,y+1,0), Vec3(x+1,y+1,0)))

        # if verts:
        self.selection_renderer.model.vertices = [v+Vec3(-.5,-.5,0) for v in verts]
        self.selection_renderer.model.triangles = [(i, i+1) for i in range(0, len(verts), 2)]
        self.selection_renderer.model.generate()


class PixelEditor(GridEditor):
    def __init__(self, texture, palette=(color.black, color.white, color.light_gray, color.gray, color.red, color.orange, color.yellow, color.lime, color.green, color.turquoise, color.cyan, color.azure, color.blue, color.violet, color.magenta, color.pink), **kwargs):
        super().__init__(texture=texture, size=texture.size, palette=palette, **kwargs)
        self.set_texture(texture)

    def set_texture(self, texture, render=True, clear_undo_stack=True):
        self.canvas.texture = texture
        self.w, self.h = int(texture.size[0]), int(texture.size[1])
        self.canvas.scale_x = self.canvas.scale_y * self.w / self.h
        self.grid = [[texture.get_pixel(x,y) for y in range(texture.height)] for x in range(texture.width)]
        self.canvas.texture.filtering = None
        self.cursor.scale = Vec2(self.brush_size / self.w, self.brush_size / self.h)
        self.help_icon.scale = self.help_icon.target_scale
        # self.selection_mover.world_scale = .5

        if clear_undo_stack:
            self.undo_stack.clear()
            self.undo_index = -1
        self.record_undo()

        if render:
            self.render()


    def draw(self, x, y):
        for _y in range(y, min(y+self.brush_size, self.h)):
            for _x in range(x, min(x+self.brush_size, self.w)):
                self.grid[_x][_y] = self.selected_char
                self.canvas.texture.set_pixel(_x, _y, self.grid[_x][_y])

        self.canvas.texture.apply()


    def render(self):
        for y in range(self.h):
            for x in range(self.w):
                self.canvas.texture.set_pixel(x, y, self.grid[x][y])

        self.canvas.texture.apply()


    def save(self):
        if self.canvas.texture.path:
            self.canvas.texture.save(self.canvas.texture.path)
            print('saved:', self.canvas.texture.path)

    @property
    def texture(self):
        return self.canvas.texture

    @texture.setter
    def texture(self, value):
        self.canvas.texture = value



class ASCIIEditor(GridEditor):
    def __init__(self, size=(61,28), palette=(' ', '#', '|', 'A', '/', '\\', 'o', '_', '-', 'i', 'M', '.'), font='VeraMono.ttf', canvas_color=color.black, line_height=1.1, **kwargs):
        super().__init__(size=size, palette=palette, canvas_color=canvas_color, **kwargs)
        rotated_grid = list(zip(*self.grid[::-1]))
        text = '\n'.join([''.join(reversed(line)) for line in reversed(rotated_grid)])

        self.text_entity = Text(parent=self.parent, text=text, x=-.0, y=.5, line_height=line_height, font=font)

        self.scale = (self.text_entity.width, self.text_entity.height)
        self.canvas.scale = 1
        self.text_entity.world_parent = self
        self.text_entity.y = 1
        self.text_entity.z = -.001

    def render(self):
        rotated_grid = list(zip(*self.grid[::-1]))
        self.text_entity.text = '\n'.join([''.join(reversed(line)) for line in reversed(rotated_grid)])


    def input(self, key):
        super().input(key)
        if held_keys['control'] and key == 'c':
            print(self.text_entity.text)
            pyperclip.copy(self.text_entity.text)
        #
        # if held_keys['control'] and key == 'v' and pyperclip.paste().count('\n') == (h-1):
        #     t.text = pyperclip.paste()
        #     undo_index += 1
        #     undo_stack = undo_stack[:undo_index]
        #     undo_stack.append(deepcopy(grid))




if __name__ == '__main__':
    app = Ursina(borderless=False)
    '''
    pixel editor example, it's basically a drawing tool.
    can be useful for level editors and such
    here we create a new texture, but can also give it an existing texture to modify.
    '''
    from PIL import Image
    t = Texture(Image.new(mode='RGBA', size=(32,32), color=(0,0,0,1)))
    from ursina.prefabs.grid_editor import PixelEditor
    editor = PixelEditor(parent=scene, texture=load_texture('brick'), scale=10)

    camera.orthographic = True
    camera.fov = 15
    EditorCamera(rotation_speed=0)
    # editor.selection_mover.enabled = True
    # editor.selection_mover.on_click()
    # editor.temp_layer.alpha = .5

    # '''
    # same as the pixel editor, but with text.
    # '''
    # from ursina.prefabs.grid_editor import ASCIIEditor
    # ASCIIEditor(x=0, scale=.1)

    app.run()



from ursina import *; ssao_shader = Shader(language=Shader.GLSL, fragment='''
#version 140


vec3 sphere[16] = vec3[](
    vec3( 0.5381, 0.1856,-0.4319), vec3( 0.1379, 0.2486, 0.4430),
    vec3( 0.3371, 0.5679,-0.0057), vec3(-0.6999,-0.0451,-0.0019),
    vec3( 0.0689,-0.1598,-0.8547), vec3( 0.0560, 0.0069,-0.1843),
    vec3(-0.0146, 0.1402, 0.0762), vec3( 0.0100,-0.1924,-0.0344),
    vec3(-0.3577,-0.5301,-0.4358), vec3(-0.3169, 0.1063, 0.0158),
    vec3( 0.0103,-0.5869, 0.0046), vec3(-0.0897,-0.4940, 0.3287),
    vec3( 0.7119,-0.0154,-0.0918), vec3(-0.0533, 0.0596,-0.5411),
    vec3( 0.0352,-0.0631, 0.5460), vec3(-0.4776, 0.2847,-0.0271)
);


uniform sampler2D tex;
uniform sampler2D dtex;
uniform sampler2D random_texture;
uniform mat4 p3d_ViewProjectionMatrix;

in vec2 uv;
out vec4 o_color;

uniform float numsamples;
uniform float radius;
uniform float amount;
uniform float strength;
uniform float falloff;


vec3 get_normal(vec2 texcoords) {
    const vec2 offset1 = vec2(0.0, 0.001);
    const vec2 offset2 = vec2(0.001, 0.0);

    float depth = texture(dtex, texcoords).r;
    float depth1 = texture(dtex, texcoords + offset1).r;
    float depth2 = texture(dtex, texcoords + offset2).r;

    vec3 p1 = vec3(offset1, depth1 - depth);
    vec3 p2 = vec3(offset2, depth2 - depth);

    vec3 normal = cross(p1, p2);
    normal.z = -normal.z;

    return normalize(normal);
}

vec3 reconstructPosition(in vec2 uv, in float z)
{
    float x = uv.x * 2.0f - 1.0f;
    float y = (1.0 - uv.y) * 2.0f - 1.0f;
    vec4 position_s = vec4(x, y, z, 1.0f);
    mat4x4 view_projection_matrix_inverse = inverse(p3d_ViewProjectionMatrix);
    vec4 position_v = view_projection_matrix_inverse * position_s;
    return position_v.xyz / position_v.w;
}


void main() {
    float depth = texture(dtex, uv).r;
    vec3 position = reconstructPosition(uv, depth);
    vec3 normal = get_normal(uv);

    vec2 noiseScale = vec2(800.0/4.0, 600.0/4.0); // screen = 800x600
    vec3 randomVec = texture(random_texture, uv * noiseScale).xyz;
    vec3 tangent   = normalize(randomVec - normal * dot(randomVec, normal));
    vec3 bitangent = cross(normal, tangent);
    mat3 TBN       = mat3(tangent, bitangent, normal);

    vec3 random_vector = normalize((texture(random_texture, uv * 18.0 + depth + normal.xy).xyz * 2.0) - vec3(1.0)).xyz;
    float occlusion = 0.0;
    float radius = radius / depth;
    float depth_difference;
    vec3 sample_normal;
    vec3 ray;

    for(int i = 0; i < numsamples; ++i) {
        vec3 random_vector = texture(random_texture, uv * i * 100).xyz;
        ray = radius * reflect(sphere[i], random_vector);
        sample_normal = get_normal(uv + ray.xy).xzy;

        depth_difference = (depth - texture(dtex, uv + ray.xy).r);
        occlusion += step(falloff, depth_difference) * (.1 - dot(sample_normal.xyz, normal)) * (1.0 - smoothstep(falloff, strength, depth_difference));
    }

    o_color.rgb = normal;
    o_color.rgb = texture(tex, uv).rgb + (occlusion * -(amount/numsamples));
    o_color.a = 1.0;
}
''',

default_input = {
    'numsamples' : 16,
    'radius' : 0.01, # 0.05 is broken and cool
    'amount' : 3.0,
    'strength' : 0.001,
    'falloff' : 0.000002,
    'random_texture' : Func(load_texture, 'noise'),
}
)
if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    e = Entity(model='sphere', color=color.orange)
    e = Entity(model='cube', y=-1)
    e = Entity(model='plane', scale=100, y=-1)
    camera.shader = ssao_shader
    #camera.clip_plane_far = 500
    camera.clip_plane_near = 1

    EditorCamera()

    def input(key):
        if key == 'space':
            if camera.shader:
                camera.shader = None
            else:
                camera.shader = ssao_shader


    random.seed(2)
    for i in range(20):
        e = Entity(model='cube', position=Vec3(random.random(),random.random(),random.random())*3, rotation=Vec3(random.random(),random.random(),random.random())*360)
        # e.shader = matcap_shader
    app.run()




class Scene(Entity):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        Entity(parent=self, position=Vec3(0.8914, 0.3707, -3.1759), rotation=Vec3(0, 40.2606, 0), scale=Vec3(5.452, 5.452, 5.452), model='wireframe_cube', origin=Vec3(0, -0.5, 0), )




class Scene(Entity):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        Entity(parent=self, position=Vec3(0.4138,0.032,-2.857), model='cube', origin=Vec3(0,-0.5,0), ignore=True)
        Entity(parent=self, position=Vec3(-3.6067,0.032,-4.2204), model='cube', origin=Vec3(0,-0.5,0), ignore=True)
        Entity(parent=self, position=Vec3(-7.2283,0.032,-2.2602), model='cube', origin=Vec3(0,-0.5,0), ignore=True)



from ursina import *


def grid_layout(l, max_x=8, max_y=8, spacing=(0,0,0), origin=(-.5,.5,0), offset=(0,0,0)):
    if len(origin) == 2:
        origin += (0,)
    if len(offset) == 2:
        offset += (0,)
    if not isinstance(l, list):
        print('error: grid_layout input must be a list or tuple, not', l.__class__.__name__)
        return
    x, y, z = 0, 0, 0

    dimensions = l[0].bounds.size
    direction = [-e*2 for e in origin]
    direction = [1 if e == 0 else e for e in direction]
    # print(direction)
    row = list()

    for i, e in enumerate(l):
        e.position = (
            x * dimensions[0] * direction[0],
            y * dimensions[1] * direction[1],
            z * dimensions[2] * direction[2]
        )
        e.position += offset
        e.origin = origin
        row.append(e)

        x += 1
        if x >= max_x:
            y += 1
            x = 0
            # center row
            if origin[0] == 0:
                for e in row:
                    e.x -= e.bounds.size.x * len(row) / 2 - e.bounds.size.x / 2
            row.clear()

        if y >= max_y:
            z += 1
            y = 0

        e.x *= 1 + spacing[0]
        e.y *= 1 + spacing[1]

    # center last row
    if origin[0] == 0:
        for e in row:
            e.x -= e.bounds.size.x * len(row) / 2 - (e.bounds.size.x / 2)

if __name__ == '__main__':
    app = Ursina()

    center = Entity(model='quad', scale=.1, color=color.red)
    p = Entity()
    for i in range(4*5):
        b = Button(parent=p, model='quad', scale=Vec2(.2,.1), text=str(i), color=color.tint(color.random_color(),-.6))
        b.text_entity.scale=1
    t = time.time()
    grid_layout(p.children, max_x=7, max_y=10, origin=(0, .5), spacing=(.15, 0))
    center = Entity(parent=camera.ui, model=Circle(), scale=.005, color=color.lime)
    EditorCamera()
    print(time.time() - t)

    # test
    for e in [(-.5,.5), (0,.5), (.5,.5), (-.5,0), (0,0), (.5,0), (-.5,-.5), (0,-.5), (.5,-.5)]:
        Button(
            text='*',
            model='quad',
            text_origin=e,
            scale=.095,
            origin=(-.5,.5),
            position = window.top_left + Vec2(*e)*.2 + Vec2(.1,-.1),
            tooltip=Tooltip(str(e)),
            on_click=Func(grid_layout, p.children, max_x=7, max_y=10, origin=e, spacing=(.15, 0))
        )



    app.run()



from panda3d.core import Texture as PandaTexture
from panda3d.core import SamplerState
from panda3d.core import Filename
from panda3d.core import PNMImage
from pathlib import Path
from ursina.vec2 import Vec2
from ursina import color
from ursina.ursinamath import clamp


class Texture():

    default_filtering = None      # options: None / 'bilinear' / 'mipmap'

    def __init__(self, value, filtering='default'):

        if isinstance(value, str):
            value = Path(value)

        if isinstance(value, Path):
            self.path = Path(value)
            self._texture = loader.loadTexture(Filename.fromOsSpecific(str(value)))
            self._cached_image = None   # for get_pixel() method

        elif isinstance(value, PandaTexture):
            self._texture = value

        else:
            from PIL import Image
            image = value
            self._texture = PandaTexture()
            self._texture.setup2dTexture(image.width, image.height, PandaTexture.TUnsignedByte, PandaTexture.FRgba)
            self._texture.setRamImageAs(image.transpose(Image.FLIP_TOP_BOTTOM).tobytes(), image.mode)
            self._cached_image = image   # for get_pixel() method
            # self._cached_image = image.transpose(Image.FLIP_TOP_BOTTOM)
            self.path = None

        if filtering == 'default':
            filtering = Texture.default_filtering      # None/'bilinear'/'mipmap' default: 'None'
        self.filtering = filtering


    @staticmethod
    def new(size, color=(255,255,255)):
        img = PNMImage(*size)
        if len(color) == 4:
            img.addAlpha()

        img.fill(*color)
        panda_tex = PandaTexture('texture')
        panda_tex.load(img)
        return Texture(panda_tex)


    @property
    def name(self):
        try:
            return self.path.name
        except:
            return f'PIL_texture_{self.size}'

    def __str__(self):
        return self.name


    @property
    def size(self):
        return Vec2(self.width, self.height)

    @property
    def width(self):
        if self._cached_image:
            return self._cached_image.size[0]
        elif self._texture.getOrigFileXSize() > 0:
            return self._texture.getOrigFileXSize()
        return 0

    @property
    def height(self):
        if self._cached_image:
            return self._cached_image.size[1]
        elif self._texture.getOrigFileYSize() > 0:
            return self._texture.getOrigFileYSize()
        return 0

    @property
    def pixels(self):
        from numpy import asarray, flip
        from PIL import Image

        if self._cached_image:
            return asarray(self._cached_image)

        pixels = asarray(Image.open(self.path))
        pixels = flip(pixels, axis=0)
        return pixels


    @property
    def filtering(self):
        return self._filtering

    @filtering.setter
    def filtering(self, value):
        # print('setting filtering:', value)
        if value in (None, False, 'nearest', 'nearest neighbor', 'point'):
            self._texture.setMagfilter(SamplerState.FT_nearest)
            self._texture.setMinfilter(SamplerState.FT_nearest)
            self._filtering = False
        elif value in (True, 'linear', 'bilinear'):
            self._texture.setMagfilter(SamplerState.FT_linear)
            self._texture.setMinfilter(SamplerState.FT_linear)
            self._filtering = True
        elif value == 'mipmap':
            self._texture.setMinfilter(SamplerState.FT_linear_mipmap_linear)
            self._filtering = 'mipmap'

    @property
    def repeat(self):
        return self._repeat

    @repeat.setter
    def repeat(self, value):
        self._repeat = value
        self._texture.setWrapU(value)
        self._texture.setWrapV(value)


    def get_pixel(self, x, y):
        try:
            if not self._cached_image:
                from PIL import Image
                self._cached_image = Image.open(self.path)


            col = self._cached_image.getpixel((x, self.height-y-1))
            if self._cached_image.mode == 'LA':
                col = (col[0], col[0], col[0], col[1])

            if self._cached_image.mode == 'L':
                col = (col[0], col[0], col[0])

            return color.rgba(*col)
        except Exception as e:
            print(e)
            return None


    def get_pixels(self, start, end):
        start = (clamp(start[0], 0, self.width), clamp(start[1], 0, self.width))
        end = (clamp(end[0], 0, self.width), clamp(end[1], 0, self.width))
        pixels = []

        for y in range(start[1], end[1]):
            for x in range(start[0], end[0]):
                pixels.append(self.get_pixel(x,y))

        return pixels

    def set_pixel(self, x, y, color):
        if not self._cached_image:
            from PIL import Image
            self._cached_image = Image.open(self.path)

        self._cached_image.putpixel((x, self.height-y-1), tuple(int(e*255) for e in color))

    def apply(self):
        from PIL import Image
        if not self._cached_image:
            self._cached_image = Image.open(self.path)

        self._texture.setRamImageAs(self._cached_image.transpose(Image.FLIP_TOP_BOTTOM).tobytes(), self._cached_image.mode)
        # self._texture.setRamImageAs(self._cached_image.tobytes(), self._cached_image.mode)

    def save(self, path):
        if not self._cached_image:
            from PIL import Image
            self._cached_image = Image.open(self.path)

        self._cached_image.save(path)

    def __repr__(self):
        return self.name


    def __del__(self):
        # self._texture.releaseAll()
        del self._cached_image



if __name__ == '__main__':
    from ursina import *
    from ursina import texture_importer
    app = Ursina()
    '''
        The Texture class rarely used manually but usually instantiated
        when assigning a texture to an Entity
        texture = Texture(path / PIL.Image / panda3d.core.Texture)

        A texture file can be a .png, .jpg or .psd.
        If it's a .psd it and no compressed version exists, it will compress it automatically.
    '''
    e = Entity(model='quad', texture='brick')
    e.texture.set_pixel(0, 2, color.blue)
    e.texture.apply()
    #
    # for y in range(e.texture.height):
    #     for x in range(e.texture.width):
    #         if e.texture.get_pixel(x,y) == color.blue:
    #             print('found blue pixel at:', x, y)

    # test
    application.asset_folder = Path(r'C:\sync\high resolution images')
    e = Entity(model='quad')
    # from PIL import Image

    from ursina.prefabs.memory_counter import MemoryCounter
    MemoryCounter()
    def input(key):
        if key == 'a':
            # img = Image.open(r'C:\sync\high resolution images\tesla_city.png')
            # print('-------', img)
            # e.texture = Texture(img)
            e.texture = 'tesla_city'

        if key == 'space':
            # if not 'tesla_city' in texture_importer.imported_textures:
            #     return
            # print('del texture')
            # del texture_importer.imported_textures['tesla_city']
            # del e.texture
            # tex = e.texture._texture
            t = e.texture._texture
            # e.texture._texture = None
            e.texture = None
            t.releaseAll()
            # print(t.releaseAll())
            t.clearRamImage()
            # e.texture._texture.clear()
            # e.texture._texture = None
            # e.texture = None
            # destroy(e)
        if key == 'p':
            for key, value in texture_importer.imported_textures.items():
                print(key, value)

    e.texture = 'brick'
    # e.texture.set_pixels([e for e in e.texture.get_pixels()])
    e.texture.apply()
    tex = Texture.new((512,512), (255,0,0))

    app.run()



from ursina import *



if __name__ == '__main__':
    app = Ursina()
    window.color=color._10
    t = Text('''powered by\nursina engine\nʕ •ᴥ•ʔゝ□''', font='unifont.ttf', scale=3, origin=(0,0))
    t.appear(speed=.05)
    t.fade_out(delay=3, duration=1, curve=curve.linear)
    # Sprite(texture='default_sky', z=1, color=color.dark_gray)
    Entity(parent=camera.ui, model='quad', texture='shore', scale_x=1.777)
    Text('Shore of Shadows', origin=(0,0), scale=6, color=color._20, y=.15, resolution=300)
    app.run()



from ursina import *


class FileButton(Button):
    def __init__(self, load_menu, **kwargs):
        super().__init__(model='quad', highlight_color=color.dark_gray, scale=(.875,.025), pressed_scale=1, selected=False)
        self.load_menu = load_menu

        for key, value in kwargs.items():
            setattr(self, key, value)

        self.text_entity.scale *= .75


    def on_click(self):
        if len(self.load_menu.selection) >= self.load_menu.selection_limit and not self.selected:
            for e in self.parent.children:
                e.selected = False

        self.selected = not self.selected
        if self.selected:
            self.load_menu.address_bar.text_entity.text = '<light_gray>' + str(self.path.resolve())
        else:
            self.load_menu.address_bar.text_entity.text = '<light_gray>' + str(self.load_menu.path.resolve())


    def on_double_click(self):
        if self.path.is_dir():
            self.load_menu.path = self.path
        else:
            self.selected = True
            self.load_menu.open()


    @property
    def selected(self):
        return self._selected

    @selected.setter
    def selected(self, value):
        self._selected = value
        if value == True:
            self.color = color.azure
            self.highlight_color = color.azure
        else:
            self.color = Button.default_color

        self.load_menu.open_button.color = color.azure if self.load_menu.selection else color.dark_gray



class FileBrowser(Entity):
    def __init__(self, **kwargs):

        self.file_types = ['.*', ]
        self.start_path = Path('.').resolve()
        super().__init__(parent=camera.ui, scale=(1,1), model='quad', color=color.clear, collider='box', y=.45)

        self.return_files = True
        self.return_folders = False
        self.selection_limit = 1
        self.max_buttons = 24

        self.title_bar = Button(parent=self, scale=(.9,.035), text='<gray>Open', color=color.dark_gray, collision=False)
        self.address_bar = Button(parent=self, scale=(.8,.035), text='//', text_origin=(-.5,0), y=-.05, highlight_color=color.black)
        self.address_bar.text_entity.scale *= .75
        self.address_bar.text_entity.x = -.5 + Text.get_width(' ')
        self.address_bar.text_entity.color = color.red

        self.folder_up_button = Button(parent=self, scale=(.035,.035), texture='arrow_down', rotation_z=180, position=(-.42,-.05,-1), color=color.white, highlight_color=color.azure, on_click=self.folder_up)
        self.button_parent = Entity(parent=self)
        self.back_panel = Entity(parent=self, model='quad', collider='box', origin_y=.5, scale=(.9,(self.max_buttons*.025)+.19), color=color._32, z=.1)
        self.bg = Button(parent=self, z=1, scale=(999,999), color=color.black66, highlight_color=color.black66, pressed_color=color.black66)

        self.cancel_button = Button(parent=self, scale=(.875*.24, .05), y=(-self.max_buttons*.025)-.15, origin_x=-.5, x=-.875/2, text='Cancel', on_click=self.close)
        self.open_button = Button(parent=self, scale=(.875*.74, .05), y=(-self.max_buttons*.025)-.15, origin_x=.5, x=.875/2, text='Open', color=color.dark_gray, on_click=self.open)

        self.cancel_button_2 = Button(parent=self.title_bar, model=Circle(), world_scale=self.title_bar.world_scale_y*.75, origin_x=.5, x=.495, z=-.1, text='<gray>x', on_click=self.close)
        self.cancel_button_2.text_entity.scale *= .75

        self.can_scroll_up_indicator = Entity(parent=self, model='quad', texture='arrow_down', rotation_z=180, scale=(.05,.05), y=-.0765, z=-.1, color=color.dark_gray, enabled=False, add_to_scene_entities=False)
        self.can_scroll_down_indicator = Entity(parent=self, model='quad', texture='arrow_down', scale=(.05,.05), y=(-self.max_buttons*.025)-.104, z=-.1, color=color.dark_gray, enabled=False, add_to_scene_entities=False)

        for key, value in kwargs.items():
            setattr(self, key ,value)

        if self.enabled:
            self.path = self.start_path     # this will populate the list
            self.scroll = 0


    def input(self, key):
        if key == 'scroll down':
            if self.scroll + self.max_buttons < len(self.button_parent.children)-1:
                self.scroll += 1

        if key == 'scroll up':
            if self.scroll > 0:
                self.scroll -= 1


    @property
    def scroll(self):
        return self._scroll

    @scroll.setter
    def scroll(self, value):
        self._scroll = value

        for i, c in enumerate(self.button_parent.children):
            if i < value or i > value + self.max_buttons:
                c.enabled = False
            else:
                c.enabled = True

        self.button_parent.y = value * .025
        self.can_scroll_up_indicator.enabled = value > 0
        self.can_scroll_down_indicator.enabled = value + self.max_buttons + 1 != len(self.button_parent.children)


    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        self._path = value
        self.address_bar.text_entity.text = '<light_gray>' + str(value.resolve())


        files = [e for e in value.iterdir() if e.is_dir() or e.suffix in self.file_types or '.*' in self.file_types]
        files.sort(key=lambda x : x.is_file())  # directories first

        for i in range(len(self.button_parent.children) - len(files)):
            destroy(self.button_parent.children.pop())


        for i, f in enumerate(files):
            prefix = ' <light_gray>'
            # if f.is_dir():
            #     prefix = '<gray> <image:folder>   <light_gray>'
            # else:
            #     prefix = ' <light_gray> <image:file_icon>   <default>'
            if i < len(self.button_parent.children):
                # just update button name and path
                self.button_parent.children[i].text_entity.text = prefix + f.name
                self.button_parent.children[i].path = f

            else:
                # print('create new:', i)
                b = FileButton(
                    parent = self.button_parent,
                    text = prefix + f.name,
                    text_origin = (-.5, 0),
                    y = -i*.025 -.09,
                    load_menu = self,
                    path = f,
                    add_to_scene_entities = False
                    )

        self.scroll = 0


    def on_enable(self):
        if not hasattr(self, 'path'):
            self.path = self.start_path
            invoke(setattr, self, 'scroll', 0, delay=.05)
            return

        self.scale = 1
        self.path = self.path
        self.button_parent.y = 0
        invoke(setattr, self, 'scroll', 0, delay=.1)


    def close(self):
        self.enabled = False


    def folder_up(self):
        self.path = self.path.parent


    def open(self, path=None):
        if not self.selection:
            return

        if not self.return_folders and self.selection[0].is_dir():
            self.path = self.selection[0]
            return

        if hasattr(self, 'on_submit'):
            self.on_submit(self.selection)

        self.close()


    @property
    def selection(self):
        return [c.path for c in self.button_parent.children if c.selected == True]





if __name__ == '__main__':
    app = Ursina()

    fb = FileBrowser(file_types=('.*'), enabled=True)

    def on_submit(paths):
        print('--------', paths)
        for p in paths:
            print('---', p)

    fb.on_submit = on_submit

    app.run()



from ursina import *
from ursina.prefabs.file_browser import FileBrowser


class FileBrowserSave(FileBrowser):
    def __init__(self, **kwargs):
        super().__init__()

        self.save_button = self.open_button
        self.save_button.color = color.azure
        self.save_button.text = 'Save'
        self.save_button.on_click = self.save
        self.file_name_field = InputField(
            parent = self,
            scale_x = .75,
            scale_y = self.save_button.scale_y,
            y = self.save_button.y,
            active = True,
            )
        # self.file_name_field.text_field.scale *= .6
        self.save_button.y -= .075
        self.cancel_button.y -= .075
        self.file_name_field.text_field.text = ''
        self.file_type = '' # to save as

        self.data = ''
        self.last_saved_file = None     # gets set when you save a file
        self.overwrite_prompt = WindowPanel(
            content=(
                Text('Overwrite?'),
                Button('Yes', color=color.azure, on_click=self.save),
                Button('Cancel')
            ),
            z=-1,
            popup=True,
            enabled=False)

        for key, value in kwargs.items():
            setattr(self, key ,value)



    def save(self):
        file_name = self.file_name_field.text_field.text
        if not file_name.endswith(self.file_type):
            file_name += self.file_type

        path = self.path / file_name
        # print('save:', path)
        if path.exists() and not self.overwrite_prompt.enabled:
            # print('overwrite file?')
            self.overwrite_prompt.enabled = True

        else:
            if isinstance(self.data, str):
                path.write_text(self.data)
            else:
                # print('write bytes')
                path.write_bytes(self.data)

            self.last_saved_file = path
            self.overwrite_prompt.enabled = False
            self.close()




if __name__ == '__main__':
    from ursina import *
    from ursina.prefabs.file_browser_save import FileBrowserSave

    app = Ursina()
    wp = FileBrowserSave(file_type = '.oto')


    import json
    save_data = {'level': 4, 'name':'Link'}
    wp.data = json.dumps(save_data)

    app.run()



from ursina import *
from ursina import color
import math
from ursina.vec3 import Vec3
# import numpy as np


def get_world_normals(model):
    import numpy as np
    normals = [np.array((n[0], n[2], n[1])) for n in model.normals]
    object_matrix = model.getNetTransform().getMat()
    normals = [object_matrix.xformVec(Vec3(*n)) for n in normals]
    normals = [Vec3(n[0], n[2], n[1]).normalized() for n in normals]
    return normals


def colorize(model, left=color.white, right=color.blue, down=color.red, up=color.green, back=color.white, forward=color.white, smooth=True, world_space=True, strength=1):

    if not model.normals:
        print('generating normals for', model)
        model.generate_normals(smooth=smooth)

    if world_space:
        normals = get_world_normals(model)
    else:
        normals = model.normals

    cols = list()
    prev_col = color.white
    for n in normals:
        c = lerp(down, up, ((n[1]*strength) +.5))

        if n[0] < 0:
            c = lerp(c, left, -(n[0]*strength))
        else:
            c = lerp(c, right, (n[0]*strength))

        if n[2] < 0:
            c = lerp(c, back, -(n[2]*strength))
        else:
            c = lerp(c, forward, (n[2]*strength))

        has_nan = False
        for e in c:
            if math.isnan(e):
                has_nan = True
                break

        if not has_nan:
            prev_col = c
            cols.append(c)
        else:
            cols.append(prev_col)


    model.colors = cols
    model.generate()


if __name__ == '__main__':
    app = Ursina()
    import random
    for i in range(10):
        e = Entity(model=load_model('sphere', path=application.internal_models_compressed_folder, use_deepcopy=True))
        e.position = (random.uniform(-3,3),random.uniform(-3,3),random.uniform(-3,3))
        e.rotation = (random.uniform(0,360),random.uniform(0,360),random.uniform(0,360))
        e.scale = random.uniform(1,3)
        e.model.colorize(smooth=False, world_space=True, strength=.5)


    Sky(color=color.gray)
    EditorCamera()
    app.run()



from collections import defaultdict
from enum import Enum


class Keys(Enum):
    left_mouse_down = 'left mouse down'
    left_mouse_up = 'left mouse up'
    middle_mouse_down = 'middle mouse down'
    middle_mouse_up = 'middle mouse up'
    right_mouse_down = 'right mouse down'
    right_mouse_up = 'right mouse up'
    double_click = 'double click'
    scroll_up = 'scroll up'
    scroll_down = 'scroll down'
    left_arrow = 'left arrow'
    left_arrow_up = 'left arrow up'
    up_arrow = 'up arrow'
    up_arrow_up = 'up arrow up'
    down_arrow = 'down arrow'
    down_arrow_up = 'down arrow up'
    right_arrow = 'right arrow'
    right_arrow_up = 'right arrow up'
    left_control = 'left control'
    right_control = 'right control'
    left_shift = 'left shift'
    right_shift = 'right shift'
    left_alt = 'left alt'
    right_alt = 'right alt'
    left_control_up = 'left control up'
    right_control_up = 'right control up'
    left_shift_up = 'left shift up'
    right_shift_up = 'right shift up'
    left_alt_up = 'left alt up'
    right_alt_up = 'right alt up'
    page_down = 'page down'
    page_down_up = 'page down up'
    page_up = 'page up'
    page_up_up = 'page up up'
    enter = 'enter'
    backspace = 'backspace'
    escape = 'escape'
    tab = 'tab'

    gamepad_left_stick_x = 'gamepad left stick x'   # held_keys only
    gamepad_left_stick_y = 'gamepad left stick y'   # held_keys only
    gamepad_right_stick_x = 'gamepad right stick x'   # held_keys only
    gamepad_right_stick_y = 'gamepad right stick y'   # held_keys only
    gamepad_left_trigger = 'gamepad left trigger'   # held_keys only
    gamepad_right_trigger = 'gamepad right trigger'   # held_keys only
    gamepad_a = 'gamepad a'
    gamepad_a_up = 'gamepad a up'
    gamepad_b = 'gamepad b'
    gamepad_b_up = 'gamepad b up'
    gamepad_x = 'gamepad x'
    gamepad_x_up = 'gamepad x up'
    gamepad_y = 'gamepad y'
    gamepad_y_up = 'gamepad y up'
    gamepad_left_stick = 'gamepad left stick'
    gamepad_left_stick_up = 'gamepad left stick up'
    gamepad_right_stick = 'gamepad right stick'
    gamepad_right_stick_up = 'gamepad right stick up'
    gamepad_back = 'gamepad back'
    gamepad_back_up = 'gamepad back up'
    gamepad_start = 'gamepad start'
    gamepad_dpad_down = 'gamepad dpad down'
    gamepad_dpad_down_up = 'gamepad dpad down up'
    gamepad_dpad_up = 'gamepad dpad up'
    gamepad_dpad_up_up = 'gamepad dpad up up'
    gamepad_dpad_left = 'gamepad dpad left'
    gamepad_dpad_left_up = 'gamepad dpad left up'
    gamepad_dpad_right = 'gamepad dpad right'
    gamepad_dpad_right_up = 'gamepad dpad right up'
    gamepad_left_shoulder = 'gamepad left shoulder'
    gamepad_left_shoulder_up = 'gamepad left shoulder up'
    gamepad_right_shoulder = 'gamepad right shoulder'
    gamepad_right_shoulder_up = 'gamepad right shoulder up'



    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        # overridden __eq__ to allow for both str and InputEvent comparisons
        if isinstance(other, Keys):
            return self.value == other.value
        return self.value == other



held_keys = defaultdict(lambda: 0)
rebinds = dict()


def bind(original_key, alternative_key):
    rebinds[original_key] = alternative_key
    if ' mouse ' in alternative_key:
        rebinds[original_key + ' up'] = alternative_key[:-5] + ' up'
        return

    rebinds[original_key + ' hold'] = alternative_key + ' hold'
    rebinds[original_key + ' up'] = alternative_key + ' up'


def unbind(key):
    if key in rebinds:
        del rebinds[key]
        del rebinds[key + ' hold']
        del rebinds[key + ' up']
    else:
        rebinds[key] = 'none'


def rebind(to_key, from_key):
    unbind(to_key)
    bind(to_key, from_key)


def input(key):
    if key.endswith('hold') or key == Keys.scroll_down or key == Keys.scroll_up:
        return

    key = key.replace('mouse down', 'mouse')

    if key.endswith(' up') and key != 'page up' and key != 'gamepad dpad up':
        held_keys[key[:-3]] = 0
    else:
        held_keys[key] = 1


def get_combined_key(key):
    '''
    Adds control, shift and alt prefix to key.
    Example: holding control and pressing 'f' would result in 'control+f'.
    This makes it easier to check for a specific combination without manually
    checking each combination of held_keys['control'], held_keys['shift'] and held_keys['alt'].
    '''

    return ''.join(e+'+' for e in ('control', 'shift', 'alt') if held_keys[e] and not e == key) + key



if __name__ == '__main__':
    from ursina import *
    from ursina import Ursina, input_handler

    app = Ursina()
    input_handler.rebind('z', 'w')  # 'z'-key will now be registered as 'w'-key

    def test():
        print('----')
    # input_handler.rebind('a', 'f')
    def input(key):
        print(key)
        if key == 'left mouse down':
            print('pressed left mouse button')

        if key == Keys.left_mouse_down:   # same as above, but with Keys enum.
            print('pressed left mouse button')


    def update():
        for key, value in held_keys.items():
            if value != 0:
                print(key, value)



    app.run()



from ursina import *


def project_uvs(model, aspect_ratio=1, direction='forward', regenerate=False):
    uvs = list()
    if direction == 'forward':
        for v in model.vertices:
            uvs.append(((v[0]+.5) / aspect_ratio, v[1]+.5))

    elif direction == 'down':
        for v in model.vertices:
            uvs.append(((v[0]+.5) / aspect_ratio, v[2]+.5))


    model.uvs = uvs

    if regenerate:
        model.generate()




if __name__ == '__main__':
    app = Ursina()

    e = Entity(model='sphere', texture='ursina_logo')
    project_uvs(e.model)
    EditorCamera()

    app.run()



from ursina import Entity, Slider, color, Button, camera, Quad, copy


class ColorPicker(Entity):
    def __init__(self, dynamic=True, **kwargs):
        super().__init__(parent=camera.ui, **kwargs)

        self.bg = Entity(parent=self, z=.01, model=Quad(aspect=.5/.2), scale=[.5,.225], origin=[0,.5], color=color.black66)
        self.h_slider = Slider(parent=self, max=360, step=1, dynamic=dynamic, on_value_changed=self._calculate_color)
        self.h_slider.bg.texture = 'rainbow'

        self.s_slider = Slider(parent=self, max=100, step=1, dynamic=dynamic, on_value_changed=self._calculate_color)
        self.s_slider.overlay = Entity(parent=self.s_slider.bg, model=copy(self.s_slider.bg.model), z=-.01, texture='horizontal_gradient', color=color.gray)

        self.v_slider = Slider(parent=self, max=100, step=1, dynamic=dynamic, on_value_changed=self._calculate_color)
        self.v_slider.bg.color = color.black
        self.v_slider.overlay = Entity(parent=self.v_slider.bg, model=copy(self.s_slider.bg.model), z=-.01, texture='horizontal_gradient', color=color.black)

        self.a_slider = Slider(parent=self, max=100, default=100, step=1, dynamic=dynamic, on_value_changed=self._calculate_color)

        self.on_value_changed = None    # assign a function here

        for i, slider in enumerate((self.h_slider, self.s_slider, self.v_slider, self.a_slider)):
            slider.knob.model.mode = 'line'
            slider.knob.model.thickness = 2
            slider.knob.model.generate()
            slider.knob.color = color.white
            slider.bg.color = color.white
            slider.y = -.05 - (i*.03)
            slider.scale = .8
            slider.x = -.25+.05

        self.preview = Button(parent=self, scale=(.5*.84,.05), origin=[0,.5], y=slider.y-.02, color=color.white)
        self._calculate_color()

        for key, value in kwargs.items():
            setattr(self, key, value)


    def _calculate_color(self):
        self.color = color.hsv(self.h_slider.value, self.s_slider.value/100, self.v_slider.value/100, self.a_slider.value/100)
        self.preview.color = self.color
        self.preview.highlight_color = self.color
        self.preview.pressed_color = self.color
        self.s_slider.bg.color = color.hsv(self.h_slider.value, 1, 1)
        # self.s_slider.overlay.color = color.hsv(self.h_slider.value, self.s_slider.value/100, 1)
        # self.v_slider.bg.color = color.hsv(self.h_slider.value, self.s_slider.value/100, 1)

        if self.on_value_changed:
            self.on_value_changed()


if __name__ == '__main__':
    from ursina import Ursina
    app = Ursina()
    ColorPicker()

    app.run()



from ursina import *;
transition_shader = Shader(name='transition_shader', language=Shader.GLSL, fragment='''
#version 140

uniform sampler2D p3d_Texture0;
uniform vec4 p3d_ColorScale;

in vec2 uv;
out vec4 COLOR;

uniform sampler2D mask_texture;
uniform float cutoff;
uniform float smooth_size;

void main() {
    vec3 color = texture2D(p3d_Texture0, uv).rgb * p3d_ColorScale.rgb;
    float value = texture(mask_texture, uv).r;
    float alpha = smoothstep(cutoff, cutoff + smooth_size, value * max(0.0, (1.0 - smooth_size) + smooth_size));

    COLOR = vec4(color.rgb, alpha * p3d_ColorScale.a);
}

''',

default_input = {
    'cutoff' : 0,
    'smooth_size' : .25,
}
)

if __name__ == '__main__':
    from ursina import *
    app = Ursina()
    window.color = color._16

    e = Entity(model='quad', shader=transition_shader, scale=5, cutoff=0)
    # mask = load_texture('explosion_particle')
    mask = load_texture('explosion_particle_2')
    mask = load_texture('impactstrikey_02_b')
    print(mask)
    e.set_shader_input('mask_texture', mask)
    EditorCamera()

    cutoff_slider = Slider(0, 1, dynamic=True, y=-.4)
    def on_value_changed():
        e.set_shader_input('cutoff', cutoff_slider.value)
    cutoff_slider.on_value_changed = on_value_changed

    smooth_slider = Slider(0, 1, dynamic=True, y=-.45)
    def on_value_changed():
        e.set_shader_input('smooth_size', smooth_slider.value)
    smooth_slider.on_value_changed = on_value_changed

    def input(key):
        if key == 'space':
            e.cutoff = 0
            e.animate('cutoff', 1, duration=.1, curve=curve.linear, delay=.05)
        # e.scale=4
        # e.animate('scale', Vec3(7,7,7), duration=.1, curve=curve.linear, delay=.05)

    # Sprite('radial_gradient')

    app.run()



from panda3d.core import NodePath, Fog
from ursina import color


class Scene(NodePath):

    def __init__(self):
        super().__init__('scene')
        self.entities = []
        self.collidables = set()
        self._children = []


    def set_up(self):
        self.reparent_to(render)
        self.fog = Fog('fog')
        self.setFog(self.fog)
        self.fog_color = color.light_gray
        self.fog_density = 0


    def clear(self):
        from ursina.ursinastuff import destroy
        from ursina import application

        to_destroy = [e for e in self.entities if not e.eternal]
        to_keep = [e for e in self.entities if e.eternal]

        for d in to_destroy:
            try:
                destroy(d)
            except Exception as e:
                print('failed to destroy entity', e)


        self.entities = to_keep


        application.sequences.clear()


    @property
    def fog_color(self):
        return self.fog.getColor()

    @fog_color.setter
    def fog_color(self, value):
        self.fog.setColor(value)


    @property
    def fog_density(self):
        return self._fog_density

    @fog_density.setter     # set to a number for exponential density or (start, end) for linear.
    def fog_density(self, value):
        self._fog_density = value
        if isinstance(value, tuple):     # linear fog
            self.fog.setLinearRange(value[0], value[1])
        else:
            self.fog.setExpDensity(value)

    @property
    def children(self):
        return [e for e in self._children if e]     # make sure list doesn't contain destroyed entities

    @children.setter
    def children(self, value):
        self._children = value


instance = Scene()



if __name__ == '__main__':
    from ursina import *
    app = Ursina()
    # yolo = Button(name='yolo', text='yolo')
    e = Entity(model='plane', color=color.black, scale=100)
    EditorCamera()
    s = Sky()

    def input(key):
        if key == 'l':
            for e in scene.entities:
                print(e.name)

        if key == 'd':
            scene.clear()
            Entity(model='cube')

    scene.fog_density = .1          # sets exponential density
    scene.fog_density = (50, 200)   # sets linear density start and end

    app.run()



from ursina import application
from ursina.entity import Entity
from ursina import curve
from ursina.ursinastuff import invoke
from ursina.ursinastuff import destroy as _destroy

from panda3d.core import Filename


class Audio(Entity):

    volume_multiplier = .5  #

    def __init__(self, sound_file_name='', autoplay=True, auto_destroy=False, **kwargs):
        super().__init__(**kwargs)
        # printvar(sound_file_name)
        if sound_file_name:
            self.clip = sound_file_name
        else:
            self.clip = None

        self.volume = 1
        self.pitch = 1
        self.balance = 0

        self.loop = False
        self.loops = 1
        self.autoplay = autoplay
        self.auto_destroy = auto_destroy

        # self.volume_variation = 0
        # self.pitch_variation = 0

        for key, value in kwargs.items():
            setattr(self, key, value)

        if autoplay:
            self.play()

        if auto_destroy:
            invoke(self.stop, destroy=True, delay=self.length)


    def __setattr__(self, name, value):
        if hasattr(self, 'clip') and self._clip:
            if name == 'volume':
                self._clip.setVolume(value * Audio.volume_multiplier)

            if name == 'pitch':
                self._clip.setPlayRate(value)

            if name == 'loop':
                self._clip.setLoop(value)

            if name == 'loops':
                self._clip.setLoopCount(value)

        try:
            super().__setattr__(name, value)
        except Exception as e:
            return e

    @property
    def clip(self):
        return self._clip

    @clip.setter
    def clip(self, value):
        if isinstance(value, str):
            self.name = value

            if '.' in value:
                file_types = ('',)
            else:
                file_types = ('.ogg', '.wav')

            for folder in (application.asset_folder, application.internal_audio_folder):
                for suffix in file_types:
                    for f in folder.glob(f'**/{value}{suffix}'):
                        p = str(f.resolve())
                        p = Filename.fromOsSpecific(p)
                        self._clip = loader.loadSfx(p)  # type: ignore
                        # print('...loaded audio clip:', f, p)
                        return

            self._clip = None
            print('no audio found with name:', value, 'supported formats: .ogg, .wav')
            return
        else:
            try:
                self._clip = value
            except Exception as e:
                print('no audio found with name:', value, 'supported formats: .ogg, .wav', e)


    @property
    def length(self):       # get the duration of the audio clip.
        return self.clip.length() if self.clip else 0

    @property
    def status(self):
        if self.clip:
            return self.clip.status()

    @property
    def ready(self):
        return 1 if self.clip and self.status > 0 else 0

    @property
    def playing(self):
        return 1 if self.clip and self.status == 2 else 0

    @property
    def time(self):
        return self.clip.get_time()

    @time.setter
    def time(self, value):
        self.clip.set_time(value)

    @property
    def balance(self):      # pan the audio. should be a value between -.5 and .5. default: 0
        return self._balance

    @balance.setter
    def balance(self, value):
        self._balance = value
        self.clip.setBalance(value*2)


    def play(self, start=0):
        if hasattr(self, 'clip') and self.clip:
            # print('play from:', start)
            self.time = start
            self.clip.play()
        else:
            print(self, 'has no audio clip')

    def pause(self):
        if self.clip:
            self.paused_volume = self.volume
            self.paused_pitch = self.pitch
            self.paused_balance = self.balance
            self.temp_time = self.time
            self.pitch = 0
            self.stop(destroy=False)

    def resume(self):
        if self.clip and hasattr(self, 'temp_time'):
            self.clip = self.name
            self.volume = self.paused_volume
            self.pitch = self.paused_pitch
            self.balance = self.paused_balance
            self.play(self.temp_time)

    def stop(self, destroy=True):
        if self.clip:
            self.clip.stop()
        if destroy:
            _destroy(self)

    def fade(self, value, duration=.5, delay=0, curve=curve.in_expo, resolution=None, interrupt=True):
        self.animate('volume', value=value, duration=duration, delay=delay, curve=curve, resolution=resolution, interrupt=interrupt)

    def fade_in(self, value=1, duration=.5, delay=0, curve=curve.in_expo, resolution=None, interrupt='finish',
                destroy_on_ended=False):
        if duration <= 0:
            self.volume = value
        else:
            self.animate('volume', value, duration=duration, delay=delay, curve=curve, resolution=resolution, interrupt=interrupt)
        if destroy_on_ended:
            _destroy(self, delay=duration + .01)

    def fade_out(self, value=0, duration=.5, delay=0, curve=curve.in_expo, resolution=None, interrupt='finish',
                 destroy_on_ended=True):
        if duration <= 0:
            self.volume = value
        else:
            self.animate('volume', value, duration=duration, delay=delay, curve=curve, resolution=resolution, interrupt=interrupt)
        if destroy_on_ended:
            _destroy(self, delay=duration + .05)


if __name__ == '__main__':
    from ursina import Ursina, printvar

    app = Ursina()
    # a = Audio('life_is_currency_wav', pitch=1)
    def input(key):
        if key == 'space':
            a = Audio('life_is_currency', pitch=random.uniform(.5,1), loop=True, autoplay=True)
    # print(a.clip)
    # a.volume=0
    # b = Audio(a.clip)
    # a2 = Audio(clip=a.clip)
    # a2 = duplicate(a)
    # a2.clip = a.clip
    # a2.play()
    # print(a2.clip)
    # a.fade_out(delay=1)
    # DebugMenu(a)

    # def input(key):
    #     if key == 'f':
    #         a.fade_out(duration=4, curve=curve.linear)
    #
    # def update():
    #     print(a.time)

    app.run()
